"""Helper functions for writing robust heuristics."""

from typing import cast, Dict, Optional, Tuple, TYPE_CHECKING
import numpy as np
from scipy import fft, signal

from ..common import Model, TX, TY
from ..utils import Array, ArrayLike


if TYPE_CHECKING:
    num_values = float
    num_samples = float
    num_spectrum_pts = float
    num_spectrum_samples = float


def param_min_sqrs(
    model: Model,
    x: TX,
    y: TY,
    scanned_param: str,
    scanned_param_values: ArrayLike["num_values", np.float64],
    defaults: Optional[Dict[str, float]] = None,
) -> Tuple[float, float]:
    """Scans one model parameter while holding the others fixed to find the value
    that gives the best fit to the data (in the minimum sum-squared residuals sense).

    A limitation of this heuristic is that all other parameters must already have a
    known value (fixed value, user estimate, heuristic or via the ``defaults``
    argument).

    This heuristic supports arbitrary numbers of x- and y-axis dimensions.

    :param x: x-axis data
    :param y: y-axis data
    :param scanned_param: name of parameter to optimize
    :param scanned_param_values: array of scanned parameter values to test
    :param defaults: optional dictionary of fallback values to use for non-scanned
       parameters, which don't have an in initial value (fixed value, user estimate or
       heuristic) set
    :returns: tuple with the value from ``scanned_param_values`` which gives the
      best fit and the root-sum-squared residuals for that value ("cost").
    """
    defaults = cast(dict, defaults or {})
    if not set(defaults.keys()).issubset(model.parameters.keys()):
        raise ValueError("Defaults must be a subset of the model parameters")
    if scanned_param in defaults.keys():
        raise ValueError("The scanned parameter cannot have a default")

    missing_values = [
        param_name
        for param_name, param_data in model.parameters.items()
        if not param_data.has_initial_value()
        and param_name != scanned_param
        and param_name not in defaults.keys()
    ]
    if any(missing_values):
        raise ValueError(f"No initial value specified for parameters: {missing_values}")

    param_values = {
        param: param_data.get_initial_value(default=defaults.get(param))
        for param, param_data in model.parameters.items()
        if param != scanned_param
    }

    scanned_param_values = np.asarray(scanned_param_values).squeeze()
    costs = np.zeros_like(scanned_param_values)
    for idx, value in np.ndenumerate(scanned_param_values):
        param_values[scanned_param] = value
        y_params = model.func(x, param_values)
        costs[idx] = np.sqrt(np.sum(np.square(y - y_params)))

    # handle a quirk of numpy indexing if only one value is passed in
    if scanned_param_values.size == 1:
        return float(scanned_param_values), float(costs)

    opt = np.argmin(costs)
    return float(scanned_param_values[opt]), float(costs[opt])


def get_sym_x(x: TX, y: TY) -> float:
    """Returns ``x_0`` such that ``y(x-x_0)`` is maximally symmetric.

    This heuristic does not require any model parameters to have value estimates.

    This heuristic supports arbitrary numbers y-axis dimensions, but only a single
    x-axis dimension.

    Limitations of the current implementation:

    * it can struggle with datasets which have significant amounts of data "in the
      wings" of the model function. Because it doesn't know anything about the model
      function (to avoid needing estimates for parameters) it can't tell if the symmetry
      point it has found is really just an area of the distribution which is
      featureless. This could potentially be fixed by changing how we divide the data
      up into "windows"
    * it currently assumes the data is on a roughly regularly sampled grid

    :param x: x-axis data
    :param y: y-axis data
    :returns: the value of ``x`` about which ``y`` is maximally symmetric
    """
    y = np.atleast_2d(y)
    x_span = x.ptp()
    num_samples = x.size
    window_min = min(x) + 0.125 * x_span
    window_max = max(x) - 0.125 * x_span

    # min of three points per window
    window_min = max(window_min, x[2])
    window_max = min(window_max, x[-3])

    window = np.argwhere(np.logical_and(x >= window_min, x <= window_max)).squeeze()
    costs = np.zeros_like(window, dtype=float)

    for point_num, idx in np.ndenumerate(window.squeeze()):
        samples_left = idx
        samples_right = num_samples - idx - 1
        num_samples_idx = min(samples_left, samples_right)
        idx_left = idx - num_samples_idx
        idx_right = idx + num_samples_idx

        samples_left = y[:, idx_left:idx]
        samples_right = y[:, (idx + 1) : (idx_right + 1)]

        assert samples_left.shape == samples_right.shape

        diff = samples_left - np.fliplr(samples_right)

        # give more weight to windows with more structure
        mu = np.mean(samples_left)
        var = np.sum(np.square(samples_left - mu)) / samples_left.size

        cost = np.sqrt(np.sum(np.square(diff))) / samples_left.size
        costs[point_num] = cost / var

    return x[window[np.argmin(costs)]]


def find_x_offset_fft(
    x: TX,
    omega: ArrayLike[("num_spectrum_pts",), np.float64],
    spectrum: ArrayLike[("num_spectrum_pts",), np.float64],
    omega_cut_off: float,
) -> float:
    """Finds the x-axis offset of a dataset from the phase of an FFT.

    This function uses the FFT shift theorem to extract the offset from the phase
    slope of an FFT.

    This heuristic does not require any model parameters to have value estimates. It has
    good noise robustness (e.g. it's not thrown off by a small number of outliers). It
    is generally accurate when the dataset is roughly regularly sampled along ``x``,
    when the model is roughly zero outside of the dataset and when the model's spectral
    content is below the dataset's Nyquist frequency.

    This heuristic supports datasets with a single x- and y-axis dimension.

    See also :func:`get_spectrum`.

    :param omega: FFT frequency axis
    :param spectrum: complex FFT data
    :param omega_cut_off: highest value of omega to use in offset estimation
    :returns: an estimate of the x-axis offset
    """
    if spectrum.ndim != 1:
        raise ValueError(
            f"Function only takes 1 y-axis dimension, not {spectrum.shape[1]}"
        )

    keep = omega < omega_cut_off
    if np.sum(keep) < 2:
        raise ValueError("Insufficient data below cut-off")

    omega = omega[keep]
    phi = np.unwrap(np.angle(spectrum[keep]))
    phi -= phi[0]

    p = np.polyfit(omega, phi, deg=1)

    x0 = min(x) - p[0]
    x0 = x0 if x0 > min(x) else x0 + x.ptp()
    return x0


def find_x_offset_sym_peak_fft(
    model: Model,
    x: TX,
    y: TY,
    omega: ArrayLike[("num_spectrum_pts",), np.float64],
    spectrum: ArrayLike[("num_spectrum_pts",), np.float64],
    omega_cut_off: float,
    test_pts: Optional[ArrayLike[("num_values",), np.float64]] = None,
    x_offset_param_name: str = "x0",
    y_offset_param_name: str = "y0",
    defaults: Optional[Dict[str, float]] = None,
):
    """Finds the x-axis offset for symmetric, peaked (maximum deviation from the
    baseline occurs at the symmetry point) functions.

    This heuristic combines the following heuristics, picking the one that gives the
    best fit (in the "lowest sum squared residuals" sense):

    A limitation of this heuristic is that all other parameters must already have a
    known value (fixed value, user estimate, heuristic or via the ``defaults``
    argument).

    This heuristic supports datasets with a single x- and y-axis dimension.

    * :func:`find_x_offset_fft`
    * Tests all points in the top quartile of deviation from the baseline
    * Optionally, user-provided "test points", taken from another heuristic. This
      allows the developer to combine the general-purpose heuristics here with
      other heuristics which make use of more model-specific assumptions

    :param model: the fit model
    :param x: x-axis data
    :param y: y-axis data
    :param omega: FFT frequency axis
    :param spectrum: complex FFT data
    :param omega_cut_off: highest value of omega to use in offset estimation
    :param test_pts: optional array of x-axis points to test
    :param x_offset_param_name: name of the x-axis offset model parameter
    :param y_offset_param_name: name of the y-axis offset model parameter
    :param defaults: optional dictionary of fallback values to use for parameters with
        no initial value specified
    :returns: an estimate of the x-axis offset
    """
    defaults = defaults or {}
    x = np.array(x)
    y = np.array(y)

    if y.squeeze().ndim != 1:
        raise ValueError(
            f"{y.shape[0]} y-axis dimensions were provided to a method which takes 1."
        )

    x0_candidates = np.array([])

    if test_pts is not None:
        x0_candidates = np.append(x0_candidates, test_pts)

    try:
        fft_candidate = find_x_offset_fft(
            x=x, omega=omega, spectrum=spectrum, omega_cut_off=omega_cut_off
        )
        x0_candidates = np.append(x0_candidates, fft_candidate)
    except ValueError:
        pass

    y0_default = defaults.get(y_offset_param_name)
    y0 = model.parameters[y_offset_param_name].get_initial_value(y0_default)
    deviations = np.argsort(np.abs(y - y0))
    top_quartile_deviations = deviations[int(len(deviations) * 3 / 4) :]
    deviations_candidates = x[top_quartile_deviations]
    x0_candidates = np.append(x0_candidates, deviations_candidates)

    best_x0, _ = param_min_sqrs(
        model=model,
        x=x,
        y=y,
        scanned_param=x_offset_param_name,
        scanned_param_values=x0_candidates,
        defaults=defaults,
    )
    return best_x0


def get_spectrum(
    x: TX,
    y: TY,
    trim_dc: bool = False,
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns the frequency spectrum (Fourier transform) of a dataset.

    NB the returned spectrum will only match the continuos Fourier transform of the
    model function in the limit where the model function is zero outside of the sampling
    window.

    NB for narrow-band signals the peak amplitude depends on where the signal frequency
    lies compared to the frequency bins.

    This function supports datasets with a single x- and y-axis dimension.

    :param x: 1D ndarray of shape (num_samples,) containing x-axis data
    :param y: 1D ndarray of shape (num_samples,) containing y-axis data
    :param trim_dc: if `True` we do not return the DC component.
    :returns: tuple of (angular freq, fft)
    """
    x = np.squeeze(x)
    y = np.squeeze(y)

    if x.ndim != 1 or y.ndim != 1:
        raise ValueError(
            "get_pgram only supports datasets with a single x-axis / y-axis dimension."
        )

    n = x.size
    dx = x.ptp() / n
    omega = np.fft.fftfreq(n, dx) * (2 * np.pi)
    y_f = fft.rfft(y) * dx

    y_f = y_f[: int(n / 2)]
    omega = omega[: int(n / 2)]

    if trim_dc:
        omega = omega[1:]
        y_f = y_f[1:]

    return omega, y_f


def get_pgram(
    x: Array[("num_samples",), np.float64],
    y: Array[("num_samples",), np.float64],
) -> Tuple[
    Array[("num_spectrum_samples",), np.float64],
    Array[("num_spectrum_samples",), np.float64],
]:
    """Returns a Lombe-Scargle periodogram for a dataset, converted into amplitude
    units.

    This function supports datasets with a single x- and y-axis dimension.

    :param x: x-axis data
    :param y: y-axis data
    :returns: tuple with the frequency axis (angular units) and the periodogram
    """
    x = np.squeeze(x)
    y = np.squeeze(y)

    if x.ndim != 1 or y.ndim != 1:
        raise ValueError(
            "get_pgram only supports datasets with a single x-axis / y-axis dimension."
        )

    dx = np.min(np.diff(x))
    n = int(x.ptp() / dx)
    f_sample = 1 / dx
    f_max = (0.5 - 1 / n) * f_sample
    df = f_sample / n

    omega_list = 2 * np.pi * np.linspace(df, f_max, n)
    pgram = signal.lombscargle(x, y, omega_list, precenter=True)
    pgram = np.sqrt(np.abs(pgram) * 4 / len(y))

    return omega_list, pgram


def find_x_offset_sampling(
    model: Model,
    x: TX,
    y: TY,
    width: float,
    x_offset_param_name: str = "x0",
) -> float:
    """Finds the x-axis offset of a dataset by stepping through a range of potential
    offset values and picking the one that gives the lowest residuals.

    This function takes a more brute-force approach by evaluating the model at a
    range of offset values, picking the one that gives the lowest residuals. This
    may be appropriate where one needs the estimate to be highly robust in the face
    of noisy, irregularly sampled data.

    :param x: x-axis data
    :param y: y-axis data
    :param width: width of the feature we're trying to find (e.g. FWHMH). Used to
        pick the spacing between offset values to try.
    :param x_offset_param_name: name of the x-axis offset parameter

    :returns: an estimate of the x-axis offset
    """
    offsets = np.arange(min(x), max(x), width / 6)
    return param_min_sqrs(
        model=model,
        x=x,
        y=y,
        scanned_param=x_offset_param_name,
        scanned_param_values=offsets,
    )[0]



from typing import List, Tuple
import numpy as np

from . import heuristics
from .triangle import Triangle
from ..common import Model, ModelParameter, TX, TY
from ..normal import NormalFitter
from ..utils import scale_no_rescale


class ConeSlice(Model):
    """Slice through a cone according to::

        z = sign * sqrt( (k_x * (x - x0))**2 + (k_y * (y - y0)) ** 2)) + z0

    This model represents a slice through the cone with fixed ``y``, given by::

        z = sign(k_x) * sqrt( (k_x * (x - x0))**2 + alpha ** 2 ) + z0

    where:

      * ``alpha = k_y * (y - y0)``
      * we use the sign of ``k_x`` to set the sign for the cone

    Floating ``z0`` and ``alpha`` without a user-estimate for either may result in an
    unreliable fit.

    See :meth:`_func` for parameter details.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [False], [False]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_func=scale_no_rescale),
        z0: ModelParameter(scale_func=scale_no_rescale, fixed_to=0),
        k: ModelParameter(scale_func=scale_no_rescale),
        alpha: ModelParameter(lower_bound=0, scale_func=scale_no_rescale),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param z0: vertical offset to the cone
        :param k: slope along ``x``
        :param alpha: offset due to being off-centre in the y-axis
        """
        return np.sign(k) * np.sqrt((k * (x - x0)) ** 2 + alpha**2) + z0

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        fit = NormalFitter(x=x, y=y, model=Triangle())
        self.parameters["x0"].heuristic = fit.values["x0"]
        self.parameters["k"].heuristic = k = fit.values["k"]

        peak_value = fit.values["y0"]  # peak_value = alpha * k + z0

        if self.parameters["z0"].has_user_initial_value():
            z0 = self.parameters["z0"].get_initial_value()
            self.parameters["alpha"].heuristic = (peak_value - z0) / k
        elif self.parameters["alpha"].has_user_initial_value():
            alpha = self.parameters["alpha"].get_initial_value()
            self.parameters["z0"].heuristic = peak_value - alpha * k
        else:  # Hope one of z0 / alpha dominates the other
            _, z0_cost = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="z0",
                scanned_param_values=[peak_value],
                defaults={"alpha": 0},
            )
            _, alpha_cost = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="alpha",
                scanned_param_values=[peak_value / k],
                defaults={"z0": 0},
            )
            if z0_cost < alpha_cost:
                self.parameters["z0"].heuristic = peak_value
                self.parameters["alpha"].heuristic = 0
            else:
                self.parameters["z0"].heuristic = 0
                self.parameters["alpha"].heuristic = peak_value



from __future__ import annotations

import copy
import dataclasses
import inspect
import logging
import numpy as np
from typing import Callable, Dict, List, Optional, Tuple, Union, TYPE_CHECKING

from .utils import Array, ArrayLike, scale_undefined, TSCALE_FUN, TX_SCALE, TY_SCALE


if TYPE_CHECKING:
    num_samples = float
    num_x_axes = float
    num_y_axes = float


logger = logging.getLogger(__name__)


TX = Union[
    float,
    ArrayLike[("num_samples",), np.float64],
    ArrayLike[
        (
            "num_x_axes",
            "num_samples",
        ),
        np.float64,
    ],
]
TY = Union[
    float,
    Array[("num_samples"), np.float64],
    Array[("num_y_axes", "num_samples"), np.float64],
]


@dataclasses.dataclass
class ModelParameter:
    """Represents a model parameter.

    Attributes:
        scale_func: callable returning a scale factor which the parameter must be
            *multiplied* by if it was fitted using ``x`` and ``y`` data that has been
            *multiplied* by the given scale factors. Scale factors are used to improve
            numerical stability by avoiding asking the optimizer to work with very large
            or very small values of ``x`` and ``y``. The callable takes the x-axis and
            y-axis scale factors as arguments. A number of default scale functions are
            provided for convenience in :py:mod:`ionics_fits.utils`.
        lower_bound: lower bound for the parameter. Fitted values are guaranteed to be
            greater than or equal to the lower bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        upper_bound: upper bound for the parameter. Fitted values are guaranteed to be
            lower than or equal to the upper bound. Parameter bounds may be used by
            fit heuristics to help find good starting points for the optimizer.
        fixed_to: if not ``None``, the model parameter is fixed to this value during
            fitting instead of being floated. This value may additionally be used by
            the heuristics to help find good initial values for other model parameters.
            The value of ``fixed_to`` must lie within the bounds of the parameter.
        user_estimate: if not ``None`` and the parameter is not fixed, this value is
            used as an initial value during fitting rather than obtaining a value from
            the heuristics. This value may additionally be used by the heuristics to
            help find good initial values for other model parameters. The value of
            ``user_estimate`` must lie within the bounds of the parameter.
        heuristic: if both of ``fixed_to`` and ``user_estimate`` are ``None``, this
            value is used as an initial value during fitting. It is set by the
            :class:`~ionics_fits.common.Model`\' s
            :meth:`~ionics_fits.common.Model.estimate_parameters` method  and should not
            be set by the user.
    """

    scale_func: TSCALE_FUN
    lower_bound: float = -np.inf
    upper_bound: float = np.inf
    fixed_to: Optional[float] = None
    user_estimate: Optional[float] = None
    heuristic: Optional[float] = None
    scale_factor: Optional[float] = dataclasses.field(init=False, default=None)

    _metadata_attrs: List[str] = dataclasses.field(init=False)

    def __post_init__(self):
        self._metadata_attrs = [
            "lower_bound",
            "upper_bound",
            "fixed_to",
            "user_estimate",
            "heuristic",
        ]

    def _format_metadata(self) -> List[str]:
        data = []

        if self.lower_bound != -np.inf:
            data.append(f"lower_bound={self.lower_bound}")
        if self.upper_bound != np.inf:
            data.append(f"upper_bound={self.upper_bound}")
        if self.fixed_to is not None:
            data.append(f"fixed_to={self.fixed_to}")
        if self.user_estimate is not None:
            data.append(f"user_estimate={self.user_estimate}")
        data.append(f"scale_func={self.scale_func.__name__}")

        return data

    def __repr__(self):
        return f"<{self.__class__.__name__}(" + ",".join(self._format_metadata()) + ")>"

    def __getattribute__(self, name):
        attr = object.__getattribute__(self, name)
        if name in ["scale_factor", "_metadata_attrs"]:
            return attr

        scale_factor = self.scale_factor
        if attr is None or scale_factor is None:
            return attr

        if name in self._metadata_attrs:
            attr /= scale_factor

        return attr

    def __setattr__(self, name, value):
        scale_factor = self.scale_factor

        if None not in [scale_factor, value] and name in self._metadata_attrs:
            value *= scale_factor
        object.__setattr__(self, name, value)

    def rescale(self, x_scales: TX_SCALE, y_scales: TY_SCALE):
        r"""Rescales the parameter metadata based on the specified x and y data scale
        factors.

        Rescaling affects the values of :attr:`lower_bound`\, :attr:`upper_bound`\,
        :attr:`fixed_to`\, :attr:`user_estimate`\, and :attr:`heuristic`.

        :param x_scales: array of x-axis scale factors
        :param y_scales: array of y-axis scale factors
        """
        if self.scale_factor is not None:
            raise RuntimeError("Attempt to rescale an already rescaled model parameter")
        self.scale_factor = self.scale_func(x_scales, y_scales)

    def unscale(self):
        """Disables rescaling of the parameter metadata"""
        if self.scale_factor is None:
            raise RuntimeError(
                "Attempt to unscale a model parameter which was not rescaled."
            )
        self.scale_factor = None

    def get_initial_value(self, default: Optional[float] = None) -> float:
        """Returns the parameter's initial value.

        For fixed parameters, this is the value the parameter is fixed to. For floated
        parameters, it is the value used to seed the fit. In the latter case, the
        initial value is retrieved from the :attr:`user_estimate` if available,
        otherwise the :attr:`heuristic` is used.

        :param default: optional value to use if no other value is available
        """
        if self.fixed_to is not None:
            value = self.fixed_to
            if self.user_estimate is not None:
                raise ValueError(
                    "User estimates must not be provided for fixed parameters"
                )
        elif self.user_estimate is not None:
            value = self.user_estimate
        elif self.heuristic is not None:
            value = self.clip(self.heuristic)
        elif default is not None:
            value = self.clip(default)
        else:
            raise ValueError("No initial value specified")

        if value < self.lower_bound or value > self.upper_bound:
            raise ValueError("Initial value outside bounds.")

        return value

    def has_initial_value(self) -> bool:
        """
        Returns ``True`` if the parameter is fixed, has a user estimate or a heuristic.
        """
        values = [self.fixed_to, self.user_estimate, self.heuristic]
        return any([None is not value for value in values])

    def has_user_initial_value(self) -> bool:
        """Returns ``True`` if the parameter is fixed or has a user estimate"""
        return self.fixed_to is not None or self.user_estimate is not None

    def clip(self, value: float) -> float:
        """Clips a value to lie between the parameter's lower and upper bounds.

        :param value: value to be clipped
        :returns: clipped value
        """
        return np.clip(value, self.lower_bound, self.upper_bound)


class Model:
    """Base class for fit models.

    A model groups a function to be fitted with associated metadata (parameter names,
    default bounds etc) and heuristics. It is agnostic about the method of fitting or
    the data statistics.

    Models may be used either as part of a fit (see :class:`~ionics_fits.common.Fitter`)
    or as a standalone function (see :meth:`__call__`).

    Class Attributes
    ================

    Attributes:
        parameters: dictionary mapping parameter names to
            :class:`~ionics_fits.common.ModelParameter` s. The parameters may be
            modified to alter their properties (bounds, user estimates, etc.).
        internal_parameters: list of "internal" model parameters. These are not directly
            used during the fit, but are rescaled in the same way as regular model
            parameters. These are used, for example, by
            :py:mod:`~ionics_fits.models.transformations` models.
    """

    parameters: Dict[str, ModelParameter]
    internal_parameters: List[ModelParameter]

    def __init__(
        self,
        parameters: Optional[Dict[str, ModelParameter]] = None,
        internal_parameters: Optional[List[ModelParameter]] = None,
    ):
        r"""
        :param parameters: optional dictionary mapping parameter names to
            :class:`~ionics_fits.common.ModelParameter`\ s. This should be ``None``
            (default) if the model has a static set of parameters, in which case the
            parameter dictionary is generated from the call signature of
            :func:`~ionics_fits.common.Model._func`. The model parameters are stored as
            :attr:`~ionics_fits.common.Model.parameters` and may be modified
            after construction to change the model behaviour during fitting (e.g. to
            change the bounds, fixed parameters, etc).
        :param internal_parameters: optional list of "internal" model parameters, which
            are not exposed to the user as arguments of
            :func:`~ionics_fits.common.Model.func`. Internal parameters
            are rescaled in the same way as regular model parameters, but are not
            otherwise used by :class:`~ionics_fits.common.Model`. These are used by
            :py:mod:`~ionics_fits.models.transformations` models, which modify the
            behaviour of other models.
        """
        if parameters is None:
            spec = inspect.getfullargspec(self._func)
            for name in spec.args[2:]:
                assert isinstance(
                    spec.annotations[name], ModelParameter
                ), "Model parameters must be instances of `ModelParameter`"
            self.parameters = {
                name: copy.deepcopy(spec.annotations[name]) for name in spec.args[2:]
            }
        else:
            self.parameters = parameters
        self.internal_parameters = internal_parameters or []

    def __call__(self, x: TX, transpose_and_squeeze=False, **kwargs: float) -> TY:
        r"""Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the results.

        Example:

        .. testcode::

            import numpy as np
            from matplotlib import pyplot as plt
            from ionics_fits.models.sinusoid import Sinusoid

            model = Sinusoid()
            x = np.linspace(0, 1)
            y = model(x, True, a=1, omega=2*np.pi, phi=0, y0=0)
            plt.plot(x, y)

        :param x: x-axis data. For models with more than one x-axis, the data should
            be shaped ``(num_x_axes, num_samples)``. For models with a single x-axis
            dimension, a 1D array may be used instead.
        :param transpose_and_squeeze: if True, the results arrays are transposed and
            squeezed proior to being returned. This is intended to be used for plotting,
            since matplotlib requires different y-series to be stored as columns.
        :param \**kwargs: values for model parameters. All model parameters
            which are not :attr:`~ionics_fits.common.ModelParameter.fixed_to` a value
            must be specified. Any parameters which are not specified default to their
            fixed values.
        :returns: the model function values
        """
        args = {
            param_name: param_data.fixed_to
            for param_name, param_data in self.parameters.items()
            if param_data.fixed_to is not None
        }
        args.update(kwargs)
        y = self.func(x, args)

        if transpose_and_squeeze:
            y = y.T.squeeze()  # pytype: disable=attribute-error

        return y

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        """
        Returns a tuple of lists of bools specifying whether the model can be rescaled
        along each x- and y-axes dimension.
        """
        raise NotImplementedError

    def rescale(self, x_scales: TX_SCALE, y_scales: TY_SCALE):
        """Rescales the model parameters based on the specified x and y data scale
        factors.

        All :attr:`~ionics_fits.common.Model.parameters` and
        :attr:`~ionics_fits.common.Model.internal_parameters` are rescaled.

        :param x_scales: array of x-axis scale factors
        :param y_scales: array of y-axis scale factors
        """
        for param_name, param_data in self.parameters.items():
            if param_data.scale_func == scale_undefined:
                raise RuntimeError(
                    f"Parameter {param_name} has an undefined scale function"
                )
            try:
                param_data.rescale(x_scales, y_scales)
            except Exception as e:
                raise RuntimeError(f"Error rescaling parameter {param_name}") from e

        for param_data in self.internal_parameters:
            param_data.rescale(x_scales, y_scales)

    def unscale(self):
        """Disables rescaling of the model parameters."""
        parameters = list(self.parameters.values()) + self.internal_parameters
        for param_data in parameters:
            param_data.unscale()

    def get_num_x_axes(self) -> int:
        """Returns the number of x-axis dimensions the model has."""
        raise NotImplementedError

    def get_num_y_axes(self) -> int:
        """Returns the number of y-axis dimensions the model has."""
        raise NotImplementedError

    def clear_heuristics(self):
        r"""Clear the heuristics for all model parameters (both exposed and internal).

        This is mainly used in :py:mod:`~ionics_fits.models.transformations`\-type
        models where the parameter estimator my be run multiple times for the same model
        instance.
        """
        for param_data in self.parameters.values():
            param_data.heuristic = None
        for param_data in self.internal_parameters:
            param_data = None

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        """Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the result.

        To use the model as a function outside of a fit,
        :meth:`~ionics_fits.common.Model.__call__` generally
        provides a more convenient interface.

        Overload this to provide a model function with a dynamic set of parameters,
        otherwise prefer to override :func:`~ionics_fits.common.Model._func`.

        :param x: x-axis data
        :param param_values: dictionary of parameter values
        :returns: array of model values
        """
        x = np.atleast_2d(x)
        return self._func(x, **param_values)

    def _func(self, x: TX) -> TY:
        r"""Evaluates the model at a given set of x-axis points and with a given set of
        parameter values and returns the results.

        Overload this in preference to :func:`~ionics_fits.common.Model.func` unless the
        :class:`~ionics_fits.common.Model` takes a dynamic set of parameters.

        :class:`~ionics_fits.common.ModelParameter`\ s should be used as the type
        annotations for the parameters arguments to define the model's parameters.
        These are used in the construction to generate the
        :attr:`~ionics_fits.common.Model.parameters` dictionary::

            from ionics_fits.utils import scale_x, scale_y

            def _func(
                self,
                x,
                a: ModelParameter(lower_bound=-1., scale_func=scale_y()),
                x0: ModelParameter(fixed_to=0, scale_func=scale_x())
            ):
                ...

        :param x: x-axis data
        :returns: array of model values
        """
        raise NotImplementedError

    def estimate_parameters(self, x: TX, y: TY):
        """Set heuristic values for model parameters.

        Typically called by :class:`~ionics_fits.common.Fitter`.

        Implementations of this method must ensure that all parameters have an initial
        value set (at least one of :attr:`~ionics_fits.common.ModelParameter.fixed_to`,
        :attr:`~ionics_fits.common.ModelParameter.user_estimate` or
        :attr:`~ionics_fits.common.ModelParameter.heuristic` must not be ``None`` for
        each parameter).

        Implementations should aim to make use of all information supplied by the user
        (bounds, user estimates, fixed values) to provide the best initial guesses for
        all parameters.

        The x and y data is sorted along the x-axis dimensions and is filtered to remove
        points with non-finite x or y values and are is rescaled if supported by the
        model.

        :param x: x-axis data
        :param y: y-axis data
        """
        raise NotImplementedError

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Returns dictionaries of values and uncertainties for the derived model
        parameters (parameters which are calculated from the fit results rather than
        being directly part of the fit) based on values of the fitted parameters and
        their uncertainties.

        :param x: x-axis data
        :param y: y-axis data
        :param fitted_params: dictionary mapping model parameter names to their
            fitted values.
        :param fit_uncertainties: dictionary mapping model parameter names to
            their fit uncertainties.
        :returns: tuple of dictionaries containing the derived parameter values and
            uncertainties.
        """
        return {}, {}


class Fitter:
    r"""Base class for fitters.

    Fitters perform maximum likelihood parameter estimation on a dataset under the
    assumption of a certain model and statistics (normal, binomial, etc) and store the
    results as attributes.

    For details about the various fitter subclasses provided by ``ionics_fits``\ , see
    :ref:`fitters`.

    Usage
    =====

    Basic usage::

        import numpy as np
        from matplotlib import pyplot as plt

        from ionics_fits.models.polynomial import Line
        from ionics_fits.normal import NormalFitter

        a = 3.2
        y0 = -9

        x = np.linspace(-10, 10)
        y = a * x + y0

        fit = NormalFitter(x, y, model=Line())
        print(f"Fitted: y = {fit.values['a']:.3f} * x + {fit.values['y0']:.3f}")

        plt.plot(x, y)
        plt.plot(*fit.evaluate())
        plt.show()

    The fit may be configured by modifying the :class:`~ionics_fits.common.Model`\ 's
    :attr:`~ionics_fits.common.Model.parameters` dictionary. This allows one to:

      * set upper and lower bounds for each parameter
      * control which parameters are fixed / floated
      * provide user estimates to be used instead of the
        :class:`~ionics_fits.common.Model`\ 's heuristics

    As an example, let's fit a sinusoid, whose frequency is already known:

    .. testcode::

        import numpy as np
        from matplotlib import pyplot as plt

        from ionics_fits.models.sinusoid import Sinusoid
        from ionics_fits.normal import NormalFitter

        omega = 2 * np.pi
        model = Sinusoid()
        model.parameters["omega"].fixed_to = omega

        params = {
            "a": 2,
            "omega": omega,
            "phi": 0.5 * np.pi,
            "y0": 0,
            "x0": 0,
            "tau": np.inf,
        }

        x = np.linspace(-3, 3, 100)
        y = model(x, True, **params)

        fit = NormalFitter(x, y, model=model)
        print(f"Amplitude: dataset = {params['a']:.3f}, fit = {fit.values['a']:.3f}")
        print(f"Phase: dataset = {params['phi']:.3f}, fit = {fit.values['phi']:.3f}")

        plt.plot(*fit.evaluate(None, True), '-.o', label="fit")
        plt.plot(x, y, label="data")
        plt.grid()
        plt.legend()
        plt.show()

    .. testoutput::

        Amplitude: dataset = 2.000, fit = 2.000
        Phase: dataset = 1.571, fit = 1.571

    ``ionics_fits`` supports fitting datasets with arbitrary x-axis and y-axis
    dimensions. The :py:mod:`~ionics_fits.models.transformations` module provides a
    number of classes which allow higher-dimensional models to be constructed from
    lower-dimension models.

    Example of fitting a dataset with a 2D x-axis:

    .. testcode::

        import numpy as np
        from matplotlib import pyplot as plt

        from ionics_fits.models.multi_x import Gaussian2D
        from ionics_fits.normal import NormalFitter

        omega = 2 * np.pi  # we know the frequency
        model = Gaussian2D()

        params = {
            "x0_x0": 0,
            "x0_x1": 3,
            "sigma_x0": 2,
            "sigma_x1": 3,
            "y0": 0,
            "a": 9,
        }

        x_0_ax = np.linspace(-3, 3, 50)
        x_1_ax = np.linspace(-10, 10, 100)
        x_0_mesh, x_1_mesh = np.meshgrid(x_0_ax, x_1_ax)
        x_shape = x_0_mesh.shape

        x = np.vstack((x_0_mesh.ravel(), x_1_mesh.ravel()))
        y = model(x, **params)

        fit = NormalFitter(x, y, model=model)

        _, y_fit = fit.evaluate(x)
        y_fit = y_fit.reshape(x_shape)

        _, axs = plt.subplots(2, 1)
        axs[0].pcolormesh(x_0_mesh, x_1_mesh, y.reshape(x_shape))
        axs[1].pcolormesh(x_0_mesh, x_1_mesh, y_fit)
        axs[0].title.set_text("Model")
        axs[1].title.set_text("Fit")
        axs[0].grid()
        axs[1].grid()
        plt.show()

    As an example of fitting a dataset with a 2D y-axis, here's how to fit Rabi flopping
    on a pair of qubits. We'll assume all parameters are the same for the two qubits,
    other than the Rabi frequencies:

    .. testcode::

        import pprint
        import numpy as np
        from matplotlib import pyplot as plt

        from ionics_fits.models.rabi import RabiFlopTime
        from ionics_fits.models.transformations.repeated_model import RepeatedModel
        from ionics_fits.normal import NormalFitter

        params = {
            "omega_0": 2 * np.pi * 1,
            "omega_1": 2 * np.pi * 2,
            "delta": 0,
            "P_readout_e": 1,
            "P_readout_g": 0
        }

        rabi_model = RabiFlopTime(start_excited=False)
        model = RepeatedModel(
            model=rabi_model,
            num_repetitions=2,
            common_params=[
                param_name for param_name in rabi_model.parameters.keys()
                if param_name != "omega"
            ]
        )

        t = np.linspace(0, 3, 100)
        y = model(t, **params)

        fit = NormalFitter(t, y, model)

        pprint.pprint(fit.values)

        plt.plot(t, y.T, ".")
        plt.legend(("qubit 0", "qubit 1"))
        plt.gca().set_prop_cycle(None)
        plt.plot(*fit.evaluate(None, True))
        plt.grid()
        plt.show()

    .. testoutput::

        {'P_readout_e': 0.9999999999,
         'P_readout_g': 1e-10,
         'delta': 0.0,
         'omega_0': 6.283185307179586,
         'omega_1': 12.566370614362235,
         't_dead': 0.0,
         'tau': inf}

    Class Attributes
    ================

    Attributes:
        x: x-axis data. The input data is sorted along the x-axis dimensions and
            filtered to contain only the "valid" points where x and y are finite.
        y: y-axis data. The input data is sorted along the x-axis dimensions x and
            filtered to contain only the "valid" points where x and y are finite.
        sigma: standard errors for each point. This is stored as an array with the same
            shape as `y`.
        values: dictionary mapping model parameter names to their fitted values
        uncertainties: dictionary mapping model parameter names to their fit
            uncertainties. For sufficiently large datasets, well-formed problems and
            ignoring covariances these are the 1-sigma confidence intervals (roughly:
            there is a 1/3 chance that the true parameter values differ from their
            fitted values by more than this much)
        derived_values: dictionary mapping names of derived parameters (parameters which
            are not part of the fit, but are calculated by the model from the fitted
            parameter values) to their values
        derived_uncertainties: dictionary mapping names of derived parameters to their
            fit uncertainties
        initial_values: dictionary mapping model parameter names to the initial values
            used to seed the fit.
        model: the fit model
        free_parameters: list of names of the model parameters floated during the fit
        x_scales: the applied x-axis scale factors
        y_scales: the applied y-axis scale factors
    """

    x: TX
    y: TY
    sigma: Optional[TY]
    values: Dict[str, float]
    uncertainties: Dict[str, float]
    derived_values: Dict[str, float]
    derived_uncertainties: Dict[str, float]
    initial_values: Dict[str, float]
    model: Model
    free_parameters: List[str]
    x_scales: TX_SCALE
    y_scales: TY_SCALE

    def __init__(self, x: TX, y: TY, model: Model):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data. For models with more than one x-axis dimension, ``x``
            should be in the form ``(num_x_axes, num_samples)``.
        :param y: y-axis data.For models with more than one y-axis dimension, ``y``
            should be in the form ``(num_y_axes, num_samples)``.
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults. The
            model is (deep) copied and stored as an attribute.
        """
        self.model = copy.deepcopy(model)

        self.x = np.atleast_2d(np.array(x, dtype=np.float64, copy=True))
        self.y = np.atleast_2d(np.array(y, dtype=np.float64, copy=True))

        self.sigma = self.calc_sigma()
        self.sigma = np.atleast_2d(self.sigma) if self.sigma is not None else None

        if self.x.ndim != 2:
            raise ValueError("x-axis data must be a 1D or 2D array.")

        if self.y.ndim != 2:
            raise ValueError("y-axis data must be a 1D or 2D array.")

        if self.x.shape[1] != self.y.shape[1]:
            raise ValueError(
                "Number of samples in the x and y datasets must match "
                f"(got {self.x.shape[1]} along x and {self.y.shape[1]} along y)."
            )

        if self.x.shape[0] != self.model.get_num_x_axes():
            raise ValueError(
                f"Expected {self.model.get_num_x_axes()} x axes, "
                f"got {self.x.shape[0]}."
            )

        if self.y.shape[0] != self.model.get_num_y_axes():
            raise ValueError(
                f"Expected {self.model.get_num_y_axes()} y axes, "
                f"got {self.y.shape[0]}."
            )

        if self.sigma is not None and self.sigma.shape != self.y.shape:
            raise ValueError(
                f"Shapes of sigma and y must match (got {self.sigma.shape} and "
                f"{self.y.shape})."
            )

        valid_x = np.all(np.isfinite(self.x), axis=0)
        valid_y = np.all(np.isfinite(self.y), axis=0)
        valid_pts = np.logical_and(valid_x, valid_y)
        assert valid_pts.ndim == 1

        (valid_inds,) = np.where(valid_pts)
        sorted_inds = valid_inds[np.lexsort(np.flipud(self.x[:, valid_inds]))]

        self.x = self.x[:, sorted_inds]
        self.y = self.y[:, sorted_inds]

        if self.sigma is not None:
            self.sigma = self.sigma[:, sorted_inds]
            if np.any(self.sigma == 0) or not np.all(np.isfinite(self.sigma)):
                raise RuntimeError(
                    "Dataset contains points with zero or infinite uncertainty."
                )

        # Rescale coordinates to improve numerics (optimizers need to do things like
        # calculate numerical derivatives which is easiest if x and y are O(1)).
        rescale_xs, rescale_ys = self.model.can_rescale()

        if len(rescale_xs) != self.model.get_num_x_axes():
            raise ValueError(
                "Unexpected number of x-axis results returned from model.can_rescale: "
                f"expected {self.model.get_num_x_axes()}, got {len(rescale_xs)}"
            )

        if len(rescale_ys) != self.model.get_num_y_axes():
            raise ValueError(
                "Unexpected number of y-axis results returned from model.can_rescale: "
                f"expected {self.model.get_num_y_axes()}, got {len(rescale_ys)}"
            )

        self.x_scales = np.array(
            [
                max(np.abs(self.x[idx, :])) if rescale else 1.0
                for idx, rescale in enumerate(rescale_xs)
            ]
        )

        self.y_scales = np.array(
            [
                max(np.abs(self.y[idx, :])) if rescale else 1.0
                for idx, rescale in enumerate(rescale_ys)
            ]
        )

        # Corner-case if a y-axis dimension has values that are all 0
        self.y_scales = np.array(
            [
                y_scale if (y_scale != 0 and np.isfinite(y_scale)) else 1.0
                for y_scale in self.y_scales
            ]
        )

        self.model.rescale(self.x_scales, self.y_scales)

        x_scaled = self.x / self.x_scales[:, None]
        y_scaled = self.y / self.y_scales[:, None]

        self.model.estimate_parameters(x_scaled, y_scaled)

        for param, param_data in self.model.parameters.items():
            if not param_data.has_initial_value():
                raise RuntimeError(
                    "No fixed_to, user_estimate or heuristic specified"
                    f" for parameter `{param}`."
                )

        self.fixed_parameters = {
            param_name: param_data.fixed_to
            for param_name, param_data in self.model.parameters.items()
            if param_data.fixed_to is not None
        }
        self.free_parameters = [
            param_name
            for param_name, param_data in self.model.parameters.items()
            if param_data.fixed_to is None
        ]

        if self.free_parameters == []:
            raise ValueError("Attempt to fit with no free parameters.")

        def free_func(x: TX, *free_param_values: float) -> TY:
            """Call the model function with the values of the free parameters."""
            params = {
                param: value
                for param, value in zip(self.free_parameters, list(free_param_values))
            }
            params.update(self.fixed_parameters)
            return self.model.func(x, params)

        fitted_params, uncertainties = self._fit(
            x_scaled, y_scaled, self.model.parameters, free_func
        )
        fitted_params.update(
            {param: value for param, value in self.fixed_parameters.items()}
        )
        uncertainties.update({param: 0 for param in self.fixed_parameters.keys()})

        self.values = {
            param: value * self.model.parameters[param].scale_factor
            for param, value in fitted_params.items()
        }
        self.uncertainties = {
            param: value * self.model.parameters[param].scale_factor
            for param, value in uncertainties.items()
        }

        self.model.unscale()

        derived = self.model.calculate_derived_params(
            self.x, self.y, self.values, self.uncertainties
        )
        self.derived_values, self.derived_uncertainties = derived

        self.initial_values = {
            param: param_data.get_initial_value()
            for param, param_data in self.model.parameters.items()
        }

    def _fit(
        self,
        x: TX,
        y: TY,
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., TY],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Implementation of the parameter estimation.

        ``Fitter`` implementations must override this method to provide a fit with
        appropriate statistics.

        :param x: rescaled x-axis data, must be a 1D array
        :param y: rescaled y-axis data
        :param parameters: dictionary of rescaled model parameters
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters

        :returns: tuple of dictionaries mapping model parameter names to their fitted
            values and uncertainties.
        """
        raise NotImplementedError

    def evaluate(
        self,
        x_fit: Optional[TX] = None,
        transpose_and_squeeze=False,
    ) -> Tuple[TX, TY]:
        r"""Evaluates the model function using the fitted parameter set.

        :param x_fit: optional x-axis points to evaluate the model at. If ``None``\ , we
            use the stored value of :attr:`x`.
        :param transpose_and_squeeze: if ``True``, the results arrays are transposed and
            squeezed prior to being returned. This is intended to be used for plotting,
            since matplotlib requires different y-series to be stored as columns.

        :returns: tuple of x-axis values used and corresponding y-axis found by
            evaluating the model.
        """
        x_fit = np.atleast_2d(x_fit if x_fit is not None else self.x)
        y_fit = np.atleast_2d(self.model.func(x_fit, self.values))

        if transpose_and_squeeze:
            return x_fit.T.squeeze(), y_fit.T.squeeze()
        return x_fit, y_fit

    def residuals(self) -> TY:
        """Returns an array of fit residuals."""
        return self.y - self.evaluate()[1]

    def calc_sigma(self) -> Optional[TY]:
        """Returns an array of standard error values for each y-axis data point

        Subclasses must override this.
        """
        raise NotImplementedError



from functools import partial
from typing import Dict, List, Optional, Tuple

import numpy as np

from ..utils import param_like
from ...common import Model, ModelParameter, TX, TY


class AggregateModel(Model):
    """Model formed by combining one or more models along the y-axis to produce a new
    model, whose number of y-axis dimensions is the sum of the y-dimensionalities of the
    aggregated models.

    When aggregating a number of identical models, use a
    :class:`~ionics_fits.models.transformations.repeated_model.RepeatedModel` instead.

    Aggregate models allow multiple data sets to be fit jointly, with some parameters
    treated as "common" - their values forced to be the same for all of the aggregated
    models.

    Example usage:

    .. testcode::

        from pprint import pprint

        from ionics_fits.models.laser_rabi import LaserFlopTimeThermal
        from ionics_fits.models.transformations.aggregate_model import AggregateModel

        rsb = LaserFlopTimeThermal(start_excited=False, sideband_index=-1)
        bsb = LaserFlopTimeThermal(start_excited=False, sideband_index=+1)

        model = AggregateModel(
            models={"rsb": rsb, "bsb": bsb},
            common_params={
                param: (rsb.parameters[param], [("rsb", param), ("bsb", param)])
                for param in rsb.parameters.keys()
            },
        )

        pprint(list(model.parameters.keys()))

    .. testoutput::

        ['P_readout_e',
         'P_readout_g',
         'eta',
         'omega',
         'tau',
         't_dead',
         'delta',
         'n_bar']

    This creates an ``AggregateModel``, which models Rabi flopping on the blue and red
    sidebands of a pair of spins coupled to a motional mode starting in a thermal state.
    In this example, all parameters for the two sideband models are fit jointly.

    The first y-axis dimension (``y[0, :]``) from the ``AggregateModel`` stores the red
    sideband with the second dimension storing the blue sideband.

    At present this class only supports models with a single y-axis dimension. This
    is just because no one got around to implementing it yet rather than any
    fundamental difficulty.
    """

    def __init__(
        self,
        models: Dict[str, Model],
        common_params: Optional[
            Dict[str, Tuple[ModelParameter, List[Tuple[str, str]]]]
        ] = None,
    ):
        """
        :param models: The models to be aggregated. This should be a dictionary mapping
            model names to model instances. The model names are used as suffixes for
            names of model parameters and derived results. For example, if one of the
            aggregated models named ``model`` has a parameter ``param``, the aggregate
            model will have a parameter ``param_model``. The same applies to the
            derived results.

            The order of the models in this dictionary defines the order of the y-axis
            dimensions for the ``AggregateModel``.

            The passed-in models are considered "owned" by the AggregateModel and should
            not be used / modified elsewhere.

        :param common_params: Optional dictionary specifying "common" model parameters.
            This feature allows multiple parameters (which can be from the same or
            different models) to be fit jointly to a single value. The common parameters
            are replaced with a new parameter, which is introduced to expose the common
            value to the user.

            The parameter metadata (limits, ``fixed_to``, ``user_estimate``, etc.) from
            the new parameter replaces the metadata for all parameters bound to it.
            Metadata set on the bound parameters is disregarded.

            The ``common_params`` dictionary keys are the names of the new model
            parameters.

            The dictionary values should be tuples containing the new model template
            parameter and a list of parameters to bind to the new parameter. The bound
            parameter lists should be lists of tuples, comprised of a pair of strings
            specifying the name of the model which owns the common parameter, and the
            name of the model parameter to make common.

            The new model parameters inherit their metadata (limits etc.) from the
            template parameters, which are (deep) copied and are not modified.
        """
        self.models = models

        axes = [model.get_num_x_axes() for model in self.models.values()]
        self.num_x_axes = axes[0]
        if axes.count(self.num_x_axes) != len(axes):
            raise ValueError("All models must have the same number of x-axes")

        # make model scale functions use correct y-axis dimension
        def wrapped_scale_func(model_idx, scale_func, x_scales, y_scales):
            return scale_func(x_scales, [y_scales[model_idx]])

        for idx, model in enumerate(self.models.values()):
            model_params = list(model.parameters.values()) + model.internal_parameters
            for parameter in model_params:
                scale_func = parameter.scale_func
                wrapped_param_scale_func = partial(wrapped_scale_func, idx, scale_func)
                parameter.scale_func = wrapped_param_scale_func
                parameter.scale_func.__name__ = scale_func.__name__

        common_params = common_params or {}

        # aggregate internal parameters from all models
        internal_parameters = []
        for model in self.models.values():
            internal_parameters += model.internal_parameters

        # organise the common parameter mapping data in ways that will be useful later

        # {new_param_name: [(model_name, param_name)]}
        self.common_param_list = {
            new_param_name: [] for new_param_name in common_params.keys()
        }

        # {model_name: [model_common_params]}
        self.model_common_params: Dict[str, List[str]] = {
            model_name: [] for model_name in self.models.keys()
        }

        # {(model_name, param_name): new_param_name}
        self.common_param_map: Dict[Tuple[str, str], str] = {}

        # {new_param_name: new_param}
        new_parameters: Dict[str, ModelParameter] = {}

        for new_param_name, (template_param, bound_params) in common_params.items():
            new_parameters[new_param_name] = param_like(template_param)
            for bind in bound_params:
                bound_model_name, bound_param_name = bind

                if bound_model_name not in self.models.keys():
                    raise ValueError(
                        f"Bound model name {bound_model_name} does not match any "
                        "aggregated model"
                    )

                model_parameters = self.models[bound_model_name].parameters
                if bound_param_name not in model_parameters.keys():
                    raise ValueError(
                        f"Bound parameter name {bound_param_name} does not match any "
                        "aggregated model"
                    )

                self.common_param_list[new_param_name] = [bind]
                self.model_common_params[bound_model_name].append(bound_param_name)
                self.common_param_map[bind] = new_param_name

        # aggregate non-common parameters from all models
        parameters: Dict[str, ModelParameter] = {}
        for model_name, model in self.models.items():
            if model.get_num_y_axes() != 1:
                raise ValueError(
                    "AggregateModel currently only supports models with a single y-axis"
                    " dimension."
                )
            parameters.update(
                {
                    f"{param_name}_{model_name}": param_data
                    for param_name, param_data in model.parameters.items()
                    if param_name not in self.model_common_params[model_name]
                }
            )

        duplicates = set(new_parameters.keys()).intersection(parameters.keys())
        if duplicates:
            raise ValueError(
                "New parameter names duplicate names of existing model parameters: "
                f"{duplicates}"
            )

        parameters.update(new_parameters)

        super().__init__(parameters=parameters, internal_parameters=internal_parameters)

    def get_num_x_axes(self) -> int:
        return self.num_x_axes

    def get_num_y_axes(self) -> int:
        return len(self.models)

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        rescale_xs, rescale_ys = zip(
            *[model.can_rescale() for model in self.models.values()]
        )
        rescale_xs = np.array(rescale_xs)
        rescale_ys = np.array(rescale_ys)

        rescale_x = list(np.all(rescale_xs, axis=0))
        rescale_ys = list(np.squeeze(rescale_ys))

        return rescale_x, rescale_ys

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        x = np.atleast_2d(x)
        num_samples = x.shape[1]
        ys = np.zeros((self.get_num_y_axes(), num_samples), dtype=np.float64)
        for idx, (model_name, model) in enumerate(self.models.items()):
            model_common_params = self.model_common_params[model_name]
            model_params = {
                param_name: param_values[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }
            model_params.update(
                {
                    bound_param_name: param_values[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )

            ys[idx, :] = np.atleast_2d(model.func(x, model_params))

        return ys

    def estimate_parameters(self, x: TX, y: TY):
        for idx, (model_name, model) in enumerate(self.models.items()):
            # replace bound model parameters with new ones based on our template
            # NB we don't do this in __init__ because we want to capture subsequent
            # changes to parameter metadata
            for bound_param_name in self.model_common_params[model_name]:
                new_param_name = self.common_param_map[(model_name, bound_param_name)]
                new_param = self.parameters[new_param_name]
                model.parameters[bound_param_name] = param_like(new_param)

            model.estimate_parameters(x, y[idx])

        # use the mean value from all models as our heuristic for common params
        for new_param_name, binds in self.common_param_list.items():
            estimates = [
                self.models[model_name].parameters[param_name].get_initial_value()
                for model_name, param_name in binds
            ]
            self.parameters[new_param_name].heuristic = np.mean(estimates)

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        derived_params = {}
        derived_uncertainties = {}

        for idx, (model_name, model) in enumerate(self.models.items()):
            model_common_params = self.model_common_params[model_name]
            model_fitted_params = {
                param_name: fitted_params[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }
            model_fit_uncertainties = {
                param_name: fit_uncertainties[f"{param_name}_{model_name}"]
                for param_name in model.parameters.keys()
                if param_name not in model_common_params
            }

            model_fitted_params.update(
                {
                    bound_param_name: fitted_params[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )
            model_fit_uncertainties.update(
                {
                    bound_param_name: fit_uncertainties[
                        self.common_param_map[(model_name, bound_param_name)]
                    ]
                    for bound_param_name in model_common_params
                }
            )

            derived = model.calculate_derived_params(
                x=x,
                y=y[idx, :],
                fitted_params=model_fitted_params,
                fit_uncertainties=model_fit_uncertainties,
            )
            model_derived_params, model_derived_uncertainties = derived
            derived_params.update(
                {
                    f"{param_name}_{model_name}": value
                    for param_name, value in model_derived_params.items()
                }
            )
            derived_uncertainties.update(
                {
                    f"{param_name}_{model_name}": value
                    for param_name, value in model_derived_uncertainties.items()
                }
            )

        return derived_params, derived_uncertainties



import logging
import numpy as np
import pprint
from scipy import optimize, stats
from typing import Callable, Dict, Optional, Tuple, TYPE_CHECKING

from .common import Fitter, Model, ModelParameter, TX, TY
from .utils import Array

if TYPE_CHECKING:
    num_samples = float
    num_samples_flattened = float
    num_x_axes = float

logger = logging.getLogger(__name__)


class NormalFitter(Fitter):
    """Fitter for Normally-distributed data.

    We use least-squares fitting as a maximum-likelihood parameter estimator for
    normally distributed data. For data that is close to normal this is usually
    a pretty good approximation of a true MLE estimator.

    See :class:`~ionics_fits.common.Fitter` for further details.
    """

    def __init__(
        self,
        x: TX,
        y: TY,
        model: Model,
        sigma: Optional[TY] = None,
        curve_fit_args: Optional[Dict] = None,
    ):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data
        :param y: y-axis data
        :param sigma: optional y-axis standard deviations.
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults. The
            model is (deep) copied and stored as an attribute.
        :param curve_fit_args: optional dictionary of keyword arguments to be passed
            into ``scipy.curve_fit``.
        """
        if sigma is None:
            self.sigma = None
        else:
            self.sigma = np.array(sigma, dtype=np.float64, copy=True)

        self.curve_fit_args = {"method": "trf"}
        if curve_fit_args is not None:
            self.curve_fit_args.update(curve_fit_args)

        super().__init__(x=x, y=y, model=model)

    def _fit(
        self,
        x: TX,
        y: TY,
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., TY],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        sigma = None if self.sigma is None else self.sigma / self.y_scales[:, None]

        free_parameters = [
            param_name
            for param_name, param_data in parameters.items()
            if param_data.fixed_to is None
        ]

        p0 = [parameters[param].get_initial_value() for param in free_parameters]
        lower = [parameters[param].lower_bound for param in free_parameters]
        upper = [parameters[param].upper_bound for param in free_parameters]
        p0_dict = {
            param: parameters[param].get_initial_value() for param in free_parameters
        }

        logger.debug(
            "Starting least-squares fitting with initial parameters: "
            f"{pprint.pformat(p0_dict)}"
        )

        def fit_func(
            _: Array[
                (
                    "num_x_axes",
                    "num_samples_flattened",
                ),
                np.float64,
            ],
            *free_param_values: float,
        ) -> Array[("num_samples_flattened",), np.float64]:
            """Call the model function with the values of the free parameters."""
            return np.ravel(free_func(x, *free_param_values))

        y_flat = np.ravel(y)
        sigma_flat = None if sigma is None else np.ravel(sigma)

        p, p_cov = optimize.curve_fit(
            f=fit_func,
            xdata=y_flat,  # not used during the fit
            ydata=y_flat,
            p0=p0,
            sigma=sigma_flat,
            absolute_sigma=sigma is not None,
            bounds=(lower, upper),
            **self.curve_fit_args,
        )

        p_err = np.sqrt(np.diag(p_cov))

        p = {param: value for param, value in zip(free_parameters, p)}
        p_err = {param: value for param, value in zip(free_parameters, p_err)}

        return p, p_err

    def calc_sigma(self) -> Optional[TX]:
        """Returns an array of standard error values for each y-axis data point
        if available.
        """
        return self.sigma

    def chi_squared(self, x: TX, y: TY, sigma: TY) -> float:
        """Returns the Chi-squared fit significance for the fitted model compared to a
        given dataset as a number between 0 and 1.

        The significance gives the probability that fit residuals as large as the ones
        we observe could have arisen through chance given our assumed statistics and
        assuming that the fitted model perfectly represents the probability
        distribution.

        A value of ``1`` indicates a perfect fit (all data points lie on the fitted
        curve) a value close to ``0`` indicates super-statistical deviations of the
        dataset from the fitted model.
        """
        x = np.atleast_2d(x)
        y = np.atleast_2d(y)
        sigma = np.atleast_2d(sigma)

        if sigma.shape != y.shape:
            raise ValueError(
                f"Mismatch between shapes of sigma ({sigma.shape}) and y ({y.shape})"
            )

        n = y.size - len(self.free_parameters)

        if n < 1:
            raise ValueError(
                "Cannot calculate chi squared for fit with "
                f"{len(self.free_parameters)} floated parameters and only "
                f"{y.size} data points."
            )

        y_fit = self.model.func(x, self.values)
        chi_2 = np.sum(((y - y_fit) / sigma) ** 2)
        p = stats.chi2.sf(chi_2, n)

        return p



from typing import List, Tuple
import numpy as np

from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_x, scale_y


class Rectangle(Model):
    """Rectangle function according to::

        x <= x_l: y = y0
        x >= x_r: y = y0
        x_r > x > x_l: y0 + a

    For ``x_l = y0 = 0``, ``x_r = inf`` this is a Heaviside step function.
    """

    def __init__(self, thresh: float = 0.5):
        """
        :param thresh: threshold used to configure the parameter estimator, which
            attempts to find the edges of the rectangle by applying this threshold to
            the peak height above the baseline to determine which points are inside
            / outside the rectangle.
        """
        self.thresh = thresh
        super().__init__()

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        a: ModelParameter(scale_func=scale_y()),
        y0: ModelParameter(scale_func=scale_y()),
        x_l: ModelParameter(scale_func=scale_x()),
        x_r: ModelParameter(scale_func=scale_x()),
    ) -> TY:
        """
        :param a: rectangle height above the baseline
        :param y0: y-axis offset
        :param x_l: left transition point
        :param- x_r: right transition point
        """
        return np.where(np.logical_and(x_r > x, x > x_l), y0 + a, y0)

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        unknowns = {
            param
            for param, param_data in self.parameters.items()
            if not param_data.has_user_initial_value()
        }

        if {"x_l", "x_r"}.issubset(unknowns):
            self.parameters["y0"].heuristic = 0.5 * (y[0] + y[-1])

        elif "x_l" not in unknowns:
            x_l = self.parameters["x_l"].get_initial_value()

            if min(x) < x_l:
                self.parameters["y0"].heuristic = np.mean(y[x < x_l])
            else:
                y0 = self.parameters["y0"].heuristic = y[-1]
                self.parameters["a"].heuristic = y[0] - y0

        elif "x_r" not in unknowns:
            x_r = self.parameters["x_r"].get_initial_value()
            if max(x) > x_r:
                self.parameters["y0"].heuristic = np.mean(y[x > x_r])
            else:
                y0 = self.parameters["y0"].heuristic = y[0]
                self.parameters["a"].heuristic = y[-1] - y0

        else:
            x_l = self.parameters["x_l"].get_initial_value()
            x_r = self.parameters["x_r"].get_initial_value()

            outside = np.logical_or(x <= x_l, x >= x_r)
            inside = np.logical_and(x > x_l, x < x_r)
            self.parameters["y0"].heuristic = np.mean(y[outside])
            y0 = self.parameters["y0"].get_initial_value()
            self.parameters["a"].heuristic = np.mean(y[outside] - y0)

        y0 = self.parameters["y0"].get_initial_value()
        self.parameters["a"].heuristic = y[np.argmax(np.abs(y - y0))] - y0
        a = self.parameters["a"].get_initial_value()

        thresh = self.thresh * (y0 + (y0 + a))
        inside = (y >= thresh) if a > 0 else (y < thresh)

        if x[inside].size == 0:
            x_l = x[0]
            x_r = x[-1]
        else:
            x_l = min(x[inside])
            x_r = max(x[inside])

        self.parameters["x_l"].heuristic = x_l
        self.parameters["x_r"].heuristic = x_r



from typing import Dict, List, Tuple
import numpy as np

from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_no_rescale


class Benchmarking(Model):
    """Benchmarking success probability decay model according to::

        y = (y0 - y_inf)*p^x + y_inf

    where ``x`` is the sequence length (number of Clifford operations).

    See :meth:`_func` for parameter details.
    """

    def __init__(self, num_qubits):
        """
        :param num_qubits: The number of qubits involved in the benchmarking sequence.
        """
        super().__init__()
        self.parameters["y_inf"].fixed_to = 1 / 2**num_qubits
        self.num_qubits = num_qubits
        self.alpha = 2**num_qubits / (2**num_qubits - 1)

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [False], [False]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        p: ModelParameter(
            lower_bound=0.0, upper_bound=1.0, scale_func=scale_no_rescale
        ),
        y0: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_no_rescale,
        ),
        y_inf: ModelParameter(
            # fixed_to set to `1 / 2**num_qubits` in the constructor
            lower_bound=0,
            upper_bound=1,
            scale_func=scale_no_rescale,
        ),
    ) -> TY:
        """Fit parameters

        :param p: depolarisation parameter
        :param y0: SPAM fidelity estimate
        :param y_inf: depolarisation offset (y-axis asymptote) (fixed to ``1/2^n`` by
          default)
        """
        y = (y0 - y_inf) * p**x + y_inf
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        y = np.squeeze(y)

        self.parameters["p"].heuristic = 1.0
        self.parameters["y0"].heuristic = max(y)
        self.parameters["y_inf"].heuristic = 1 / 2**self.num_qubits

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Derived parameters:

        * ``e``: error per Clifford ``e = (1 - p) / alpha_n`` where
          ``alpha_n = 2^n / (2^n - 1)``
        * ``e_spam``: estimated SPAM error ``e_spam = 1 - y0``
        """
        p = fitted_params["p"]
        y0 = fitted_params["y0"]

        e = (1 - p) / self.alpha
        e_spam = 1 - y0

        p_err = fit_uncertainties["p"]
        y0_err = fit_uncertainties["y0"]

        derived_params = {}
        derived_params["e"] = e
        derived_params["e_spam"] = e_spam

        derived_uncertainties = {}
        derived_uncertainties["e"] = p_err / self.alpha
        derived_uncertainties["e_spam"] = y0_err

        return derived_params, derived_uncertainties



from typing import Dict, List, Tuple

from ...common import Model, ModelParameter, TX, TY


class ReparametrizedModel(Model):
    r"""Model formed by reparametrizing an existing :class:`~ionics_fits.common.Model`.

    ``ionics_fits`` aims to provide convenient and flexible model parametrisations,
    however sometimes the default parametrisation won't be convenient for your
    application. For these cases ``ReparametrizedModel``\s provide a convenient way of
    changing and extending the parametrisation of an existing model.

    Reparametrizing a model involves replacing "bound" parameters with "new" parameters,
    whose values the bound parameter values are calculated from.

    All non-bound parameters of the original model as well as all "new" parameters are
    parameters of the :class:`ReparametrizedModel` (bound parameters are internal
    parameters of the new model, but are not directly exposed to the user). All derived
    results from the original model are derived results from the
    :class:`ReparametrizedModel` (override ``calculate_derived_params`` to change
    this behaviour).

    Values and uncertainties for the bound parameters are exposed as derived results.

    Subclasses must override :meth:`bound_param_values`,
    :meth:`bound_param_uncertainties` and :meth:`bound_param_uncertainties` to specify
    the mapping between "new" and "bound" parameters.

    Example usage converting a sinusoid parameterised by offset and amplitude into one
    parametrised by minimum and maximum values:

    .. testcode::

        from typing import Dict

        from ionics_fits.models.sinusoid import Sinusoid
        from ionics_fits.models.transformations.reparametrized_model import (
            ReparametrizedModel
        )

        class SineMinMax(ReparametrizedModel):
            def __init__(self):
                super().__init__(
                    model=Sinusoid(),
                    new_params={
                        "min": ModelParameter(scale_func=scale_y()),
                        "max": ModelParameter(scale_func=scale_y()),
                    },
                    bound_params=["a", "y0"],
                )

            @staticmethod
            def bound_param_values(param_values: Dict[str, float]) -> Dict[str, float]:
                return {
                    "a": 0.5 * (param_values["max"] - param_values["min"]),
                    "y0": 0.5 * (param_values["max"] + param_values["min"]),
                }

            @staticmethod
            def bound_param_uncertainties(
                param_values: Dict[str, float], param_uncertainties: Dict[str, float]
            ) -> Dict[str, float]:
                err = 0.5 * np.sqrt(
                    param_uncertainties["max"] ** 2 + param_uncertainties["min"] ** 2
                )
                return {"a": err, "y0": err}

            @staticmethod
            def new_param_values(model_param_values: Dict[str, float]
            ) -> Dict[str, float]:
                return {
                    "max": (model_param_values["y0"] + model_param_values["a"]),
                    "min": (model_param_values["y0"] - model_param_values["a"]),
                }

    See also :class:`~ionics_fits.models.transformations.mapped_model.MappedModel`.
    """

    def __init__(
        self,
        model: Model,
        new_params: Dict[str, ModelParameter],
        bound_params: List[str],
    ):
        """
        :param model: The model to be reparametrized. This model is considered "owned"
          by the ``ReparametrizedModel`` and should not be used / modified elsewhere.
        :param new_params: dictionary of new parameters of the ``ReparametrizedModel``
        :param bound_params: list of parameters of the
          :class:`~ionics_fits.common.Model` to bound. These parameters are
          not exposed as parameters of the ``ReparametrizedModel``.
        """
        self.model = model
        self.bound_params = bound_params
        self.unbound_model_params = [
            param_name
            for param_name in self.model.parameters.keys()
            if param_name not in bound_params
        ]

        internal_parameters = self.model.internal_parameters
        parameters = dict(self.model.parameters)

        if not set(bound_params).issubset(set(parameters.keys())):
            raise ValueError("Bound parameters must all be parameters of the model")

        for param_name in bound_params:
            internal_parameters.append(parameters.pop(param_name))

        duplicates = set(new_params.keys()).intersection(parameters.keys())
        if duplicates:
            raise ValueError(
                "New parameter names must not duplicate names of unbound model "
                f"parametes. Duplicates are {duplicates}."
            )

        parameters.update(new_params)

        super().__init__(parameters=parameters, internal_parameters=internal_parameters)

    @staticmethod
    def bound_param_values(param_values: Dict[str, float]) -> Dict[str, float]:
        """Returns a dictionary of values of the model's bound parameters.

        This method must be overridden to specify the mapping from parameters of the
        ``ReparameterizedModel`` to values of the bound parameters.

        :param new_param_values: dictionary of parameter values for the
          ``ReparameterizedModel``.
        :returns: dictionary of values for the bound parameters of the original model.
        """
        raise NotImplementedError

    @staticmethod
    def bound_param_uncertainties(
        param_values: Dict[str, float], param_uncertainties: Dict[str, float]
    ) -> Dict[str, float]:
        """Returns a dictionary of uncertainties for the model's bound parameters.

        This method must be overridden to specify the mapping from parameter
        uncertainties for the ``ReparameterizedModel`` to bound parameter
        uncertainties.

        :param param_values: dictionary of values for parameters of the
          ``ReparameterizedModel``.
        :param param_uncertainties: dictionary of uncertainties for parameters of the
          ``ReparameterizedModel``.
        :returns: dictionary of values for the bound parameters of the original model.
        """
        raise NotImplementedError

    @staticmethod
    def new_param_values(model_param_values: Dict[str, float]) -> Dict[str, float]:
        r"""Returns a dictionary of values of the model's "new" parameters.

        This method must be overridden to specify the mapping from values of the
        original model to values of the ``ReparametrizedModel``\'s "new" parameters.

        This is used to find estimates for the new parameters from the original
        model's parameter estimates.

        :param model_param_values: dictionary of parameter values for the original model
        :returns: dictionary of values for the "new" parameters of the
          ``ReparametrizedModel``.
        """
        raise NotImplementedError

    def get_num_x_axes(self) -> int:
        return self.model.get_num_x_axes()

    def get_num_y_axes(self) -> int:
        return self.model.get_num_y_axes()

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return self.model.can_rescale()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        model_values = self.bound_param_values(param_values)
        model_values.update(
            {
                param_name: param_values[param_name]
                for param_name in self.unbound_model_params
            }
        )
        return self.model.func(x, model_values)

    def estimate_parameters(self, x: TX, y: TX):
        self.model.estimate_parameters(x=x, y=y)
        model_heuristics = {
            param_name: param_data.get_initial_value()
            for param_name, param_data in self.model.parameters.items()
        }
        new_heuristics = self.new_param_values(model_heuristics)

        for param_name in self.unbound_model_params:
            self.parameters[param_name].heuristic = model_heuristics[param_name]

        for param_name, param_heuristic in new_heuristics.items():
            self.parameters[param_name].heuristic = param_heuristic

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:

        bound_values = self.bound_param_values(fitted_params)
        bound_uncertainties = self.bound_param_uncertainties(
            fitted_params, fit_uncertainties
        )

        model_fitted_params = {
            param_name: fitted_params[param_name]
            for param_name in self.unbound_model_params
        }
        model_fit_uncertainties = {
            param_name: fit_uncertainties[param_name]
            for param_name in self.unbound_model_params
        }

        model_fitted_params.update(bound_values)
        model_fit_uncertainties.update(bound_uncertainties)

        derived_values, derived_uncertainties = self.model.calculate_derived_params(
            x=y,
            y=y,
            fitted_params=model_fitted_params,
            fit_uncertainties=model_fit_uncertainties,
        )

        derived_values.update(bound_values)
        derived_uncertainties.update(bound_uncertainties)

        return derived_values, derived_uncertainties



from typing import Dict, List, Tuple

from ...common import Model, TX, TY


class ScaledModel(Model):
    r"""Model with rescaled x-axis.

    A common use-case for ``ScaledModel``\s is converting models between linear and
    angular units.
    """

    def __init__(self, model: Model, x_scale: float, x_offset: float = 0.0):
        """
        :param model: model to rescale. This model is considered "owned" by the
          ``ScaledModel`` and should not be used/modified elsewhere.
        :param x_scale: multiplicative x-axis scale factor. To convert a model that
          takes x in angular units and convert to one that takes x in linear units use
          ``x_scale = 2 * np.pi``
        :param x_offset: additive x-axis offset
        """
        self.model = model
        self.x_scale = x_scale
        self.x_offset = x_offset

        super().__init__(
            parameters=self.model.parameters,
            internal_parameters=self.model.internal_parameters,
        )

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        return self.model.func(x * self.x_scale + self.x_offset, param_values)

    def estimate_parameters(self, x: TX, y: TY):
        self.model.estimate_parameters(x * self.x_scale + self.x_offset, y)
        for param_name, param_data in self.model.parameters.items():
            self.parameters[param_name].heuristic = param_data.get_initial_value()

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return self.model.can_rescale()

    def get_num_x_axes(self) -> int:
        return self.model.get_num_x_axes()

    def get_num_y_axes(self) -> int:
        return self.model.get_num_y_axes()

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        return self.model.calculate_derived_params(
            x=x * self.x_scale + self.x_offset,
            y=y,
            fitted_params=fitted_params,
            fit_uncertainties=fit_uncertainties,
        )



import copy
import inspect
from typing import Callable, Dict, Tuple, TYPE_CHECKING

import numpy as np
from scipy import special

from .quantum_phys import (
    coherent_state_probs,
    squeezed_state_probs,
    thermal_state_probs,
)
from .rabi import RabiFlop, RabiFlopFreq, RabiFlopTime
from .utils import param_like
from ..common import ModelParameter, TX, TY
from ..utils import Array, scale_invariant, scale_y, scale_undefined

if TYPE_CHECKING:
    num_fock_states = float


class LaserFlop(RabiFlop):
    r"""Base class for damped Rabi flopping with finite Lamb-Dicke parameter.

    This model calculates measurement outcomes for systems containing two internal
    states and moving in a 1D harmonic potential that undergo damped Rabi oscillations,
    defined by::

        P = P_readout_g + (P_readout_e - P_readout_g) * P_e

    where ``P_e`` is the (time-dependent) population in the excited state and
    ``P_readout_g`` and ``P_readout_e`` are the readout levels (measurement outcomes
    when the qubit is in one state).

    This class does not support fitting directly; use one of the subclasses instead.
    Subclasses must inherit from this class and a suitable
    :class:`~ionics_fits.models.rabi.RabiFlop` subclass, such as
    :class:`~ionics_fits.models.rabi.RabiFlopFreq` or
    :class:`~ionics_fits.models.rabi.RabiFlopTime`.

    The model requires that the spin state of the system starts out entirely in one
    of the ground or excited states, specified using :meth:`__init__`\'s
    ``start_excited`` parameter. It further assumes that the motional part of the system
    starts out in a distribution over different Fock states, described by the specified
    ``distribution_fun``.

    Independent variables:

        * ``t_pulse``: duration of driving pulse including dead time. The duration of
          the interaction is given by ``t = max(0, t_pulse - t_dead)``.
        * ``w``: frequency of driving pulse relative to the reference frequency
          ``w_0``\, given by ``delta = w - w_0``

    The model additionally gains any parameters associated with the specified Fock
    state distribution function.

    Derived parameters are inherited from :class:`~ionics_fits.models.rabi.RabiFlop`\'s
    :func:`~ionics_fits.models.rabi.RabiFlop.calculate_derived_params` method.

    All frequencies are in angular units.

    See also :class:`~ionics_fits.models.rabi.RabiFlop`.
    """

    def __init__(
        self,
        distribution_fun: Callable[..., Array[("num_fock_states",), np.float64]],
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        r"""
        :param distribution_fun: function returning an array of Fock state occupation
          probabilities. The distribution function's first argument should be the
          maximum Fock state to include in the simulation (the returned array has
          ``n_max + 1`` elements). Subsequent arguments should be
          :class:`~ionics_fits.common.ModelParameter`\s used to parametrise the
          distribution.
        :param start_excited: if ``True`` the qubit starts in the excited state
        :param sideband_index: change in motional state due to a pi-pulse starting
          from the spin ground-state.
        :param n_max: maximum Fock state used in the simulation
        """
        super().__init__(start_excited=start_excited)

        self.distribution_fun = distribution_fun
        self.sideband_index = sideband_index
        self.n_max = n_max

        n = np.arange(self.n_max + 1)
        if self.start_excited:
            self._n_min = n_min = np.minimum(n, n - self.sideband_index)
            n_max = np.maximum(n, n - self.sideband_index)
        else:
            self._n_min = n_min = np.minimum(n, n + self.sideband_index)
            n_max = np.maximum(n, n + self.sideband_index)

        # sqrt(n_min!/n_max!)
        self.fact = np.exp(
            0.5 * (special.gammaln(n_min + 1) - special.gammaln(n_max + 1))
        )
        self.fact[self._n_min < 0] = 0

        # Add parameters associated with the Fock state distribution function
        spec = inspect.getfullargspec(distribution_fun)
        self.distribution_params = {
            name: copy.deepcopy(spec.annotations[name]) for name in spec.args[1:]
        }

        assert all(
            [
                isinstance(param, ModelParameter)
                for param in self.distribution_params.values()
            ]
        ), "Distribution function parameters must be instances of `ModelParameter`"
        assert not any(
            [
                param_name in self.parameters.keys()
                for param_name in self.distribution_params.keys()
            ]
        ), "Distribution parameter names must not match model parameter names"
        self.parameters.update(self.distribution_params)

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        return super().func(x=x, param_values=param_values)

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: Tuple[TX, TX],
        P_readout_e: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        P_readout_g: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        eta: ModelParameter(lower_bound=0.0, scale_func=scale_invariant),
        omega: ModelParameter(lower_bound=0.0, scale_func=scale_undefined),
        tau: ModelParameter(
            lower_bound=0.0, fixed_to=np.inf, scale_func=scale_undefined
        ),
        t_dead: ModelParameter(
            lower_bound=0.0, fixed_to=0.0, scale_func=scale_undefined
        ),
        w_0: ModelParameter(scale_func=scale_undefined),
        # Fock state distribution function parameters
        **kwargs: ModelParameter,
    ) -> TY:
        """Return measurement probability.

        :param x: tuple of ``(t_pulse, w)``. Subclasses should override func to
            map this onto the appropriate input data.
        :param P_readout_e: excited state readout level
        :param P_readout_g: ground state readout level
        :param eta: Lamb-Dicke parameter
        :param omega: carrier Rabi frequency
        :param tau: decay time constant
        :param t_dead: dead time
        :param w_0: resonance frequency offset
        """
        # Rabi frequency as function of Fock state which the ion occupies
        # initially. If the ion starts in |g>, corresponds to coupling between
        # |g>|n> and |e>|n+sideband_index>. If the ion starts in |e>, corresponds
        # to coupling between |g>|n-sideband_index> and |e>|n>.
        omega_vec = (
            omega
            * np.exp(-0.5 * eta**2)
            * eta ** np.abs(self.sideband_index)
            * self.fact
            * special.eval_genlaguerre(
                self._n_min, np.abs(self.sideband_index), eta**2
            )
        )

        t_vec = np.clip(x[0] - t_dead, a_min=0, a_max=None)
        detuning_vec = x[1]
        t, detuning, omega = np.meshgrid(t_vec, detuning_vec, omega_vec, indexing="ij")

        # Transition probability for individual Fock state, given by
        #     P_transition = 1/2 * omega^2 / W^2 * [1 - exp(-t / tau) * cos(W * t)]
        W = np.sqrt(omega**2 + detuning**2)
        P_trans = (
            0.5
            * (np.divide(omega, W, out=np.zeros_like(W), where=(W != 0.0)) ** 2)
            * (1 - np.exp(-t / tau) * np.cos(W * t))
        )

        # Occupation probabilities of Fock states
        P_fock = self.distribution_fun(self.n_max, **kwargs)
        # Transition probability averaged over Fock state distribution
        P_trans_mean = np.sum(P_fock * P_trans, axis=-1).squeeze()

        P_e = 1 - P_trans_mean if self.start_excited else P_trans_mean

        return P_readout_g + (P_readout_e - P_readout_g) * P_e

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        # Pick sensible starting values which are usually good enough for the fit to
        # converge from.
        if (
            self.parameters["omega"].has_user_initial_value()
            and not self.parameters["eta"].has_user_initial_value()
        ):
            # use the rabi heuristic to find eta * omega
            eta_heuristic = True
            omega_param = self.parameters["omega"]

            self.parameters["omega"] = param_like(omega_param)
            self.parameters["eta"].heuristic = 0  # avoid exception due to no value

            omega_lower = self.parameters["omega"].lower_bound or 0
            omega_upper = self.parameters["omega"].upper_bound or np.inf

            eta_lower = self.parameters["eta"].lower_bound or 0
            eta_upper = self.parameters["eta"].upper_bound or np.inf

            self.parameters["omega"].fixed_to = None
            self.parameters["omega"].user_estimate = None
            self.parameters["omega"].lower_bound = omega_lower * eta_lower
            self.parameters["omega"].upper_bound = omega_upper * eta_upper
        else:
            eta_heuristic = False

        if "n_bar" in self.parameters.keys():
            self.parameters["n_bar"].heuristic = 1
        if "alpha" in self.parameters.keys():
            self.parameters["alpha"].heuristic = 1
        if "zeta" in self.parameters.keys():
            self.parameters["zeta"].heuristic = 1

        super().estimate_parameters(x=x, y=y)

        if eta_heuristic:
            eta_omega = self.parameters["omega"].get_initial_value()
            self.parameters["eta"].heuristic = (
                eta_omega / omega_param.get_initial_value()
            )
            self.parameters["omega"] = omega_param

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        # use eta * omega for calculating pi times etc
        fitted_params = dict(fitted_params)
        fit_uncertainties = dict(fit_uncertainties)

        omega = fitted_params.pop("omega")
        eta = fitted_params.pop("eta")

        omega_uncert = fit_uncertainties.pop("omega")
        eta_uncert = fit_uncertainties.pop("eta")

        fitted_params["omega"] = eta * omega
        fit_uncertainties["omega"] = np.sqrt(omega_uncert**2 + eta_uncert**2)

        return super().calculate_derived_params(
            x=x,
            y=y,
            fitted_params=fitted_params,
            fit_uncertainties=fit_uncertainties,
        )


class LaserFlopFreqCoherent(LaserFlop, RabiFlopFreq):
    """Fit model for Rabi flopping pulse detuning scans when the motional degree of
    freedom starts in a coherent state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=coherent_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )


class LaserFlopFreqThermal(LaserFlop, RabiFlopFreq):
    """Fit model for Rabi flopping pulse detuning scans when the motional degree of
    freedom starts in a thermal state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=thermal_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )


class LaserFlopFreqSqueezed(LaserFlop, RabiFlopFreq):
    """Fit model for Rabi flopping pulse detuning scans when the motional degree of
    freedom starts in a squeezed state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=squeezed_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )


class LaserFlopTimeCoherent(LaserFlop, RabiFlopTime):
    """Fit model for Rabi flopping pulse duration scans when the motional degree of
    freedom starts in a coherent state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=coherent_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )


class LaserFlopTimeThermal(LaserFlop, RabiFlopTime):
    """Fit model for Rabi flopping pulse duration scans when the motional degree of
    freedom starts in a thermal state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=thermal_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )


class LaserFlopTimeSqueezed(LaserFlop, RabiFlopTime):
    """Fit model for Rabi flopping pulse duration scans when the motional degree of
    freedom starts in a squeezed state.
    """

    def __init__(
        self,
        start_excited: bool,
        sideband_index: int,
        n_max: int = 30,
    ):
        super().__init__(
            distribution_fun=squeezed_state_probs,
            start_excited=start_excited,
            sideband_index=sideband_index,
            n_max=n_max,
        )



from typing import Dict, List, Optional, Tuple

from ...common import Model, TX, TY


class MappedModel(Model):
    r"""Wraps a :class:`~ionics_fits.common.Model`\, allowing parameters to be renamed
    or replaced with fixed values.

    See also
    :class:`~ionics_fits.models.transformations.reparametrized_model.ReparametrizedModel`
    """

    def __init__(
        self,
        model: Model,
        param_mapping: Dict[str, str],
        fixed_params: Optional[Dict[str, float]] = None,
        derived_result_mapping: Optional[Dict[str, Optional[str]]] = None,
    ):
        """
        :param model: The model to be wrapped. This model is considered "owned" by the
            ``MappedModel`` and should not be modified / used elsewhere.
        :param param_mapping: dictionary mapping names of parameters in the
            ``WrappedModel`` to names of parameters in the model being wrapped.
        :param fixed_params: dictionary mapping names of ``WrappedModel`` parameters
            to values they sould be fixed to. Fixed parameters are not parameters of the
            ``WrappedModel``.
        :param derived_result_mapping: optional dictionary mapping names of derived
            result in the ``WrappedModel`` to names of derived results in the model
            being wrapped. Derived results may be renamed to ``None`` to exclude them
            from the ``WrappedModel``.
        """
        self.model = model
        self.derived_result_mapping = derived_result_mapping or {}
        wrapped_params = self.model.parameters

        fixed_params = fixed_params or {}

        if unknown_mapped_params := set(param_mapping.values()) - wrapped_params.keys():
            raise ValueError(
                "The target of parameter mappings must be parameters of the inner "
                f"model. The following mapping targets are not: {unknown_mapped_params}"
            )

        if unknown_fixed_params := fixed_params.keys() - wrapped_params.keys():
            raise ValueError(
                "Fixed parameters must be parameters of the inner model. The "
                f"follow fixed parameters are not: {unknown_fixed_params}"
            )

        if missing_params := wrapped_params.keys() - (
            fixed_params.keys() | param_mapping.values()
        ):
            raise ValueError(
                "All parameters of the inner model must be either mapped of "
                "fixed. The following inner model parameters are neither: "
                f"{missing_params}"
            )

        if duplicated_params := fixed_params.keys() & param_mapping.values():
            raise ValueError(
                "Parameters cannot be both mapped and fixed. The following "
                f"parameters are both: {duplicated_params}"
            )

        self.fixed_params = {
            param_name: self.model.parameters[param_name]
            for param_name in fixed_params.keys()
        }
        for param_name, fixed_to in fixed_params.items():
            self.fixed_params[param_name].fixed_to = fixed_to

        self.param_mapping = param_mapping
        exposed_params = {
            new_name: self.model.parameters[old_name]
            for new_name, old_name in param_mapping.items()
        }
        internal_parameters = (
            list(self.fixed_params.values()) + self.model.internal_parameters
        )

        super().__init__(
            parameters=exposed_params,
            internal_parameters=internal_parameters,
        )

    def get_num_x_axes(self) -> int:
        return self.model.get_num_x_axes()

    def get_num_y_axes(self) -> int:
        return self.model.get_num_y_axes()

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return self.model.can_rescale()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        new_params = {
            old_name: param_values[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_params.update(
            {
                param_name: param_data.fixed_to
                for param_name, param_data in self.fixed_params.items()
            }
        )
        return self.model.func(x, new_params)

    def estimate_parameters(self, x: TX, y: TY):
        self.model.estimate_parameters(x, y)

        for param_name, param_data in self.parameters.items():
            old_name = self.param_mapping[param_name]
            old_param = self.model.parameters[old_name]
            param_data.heuristic = old_param.heuristic

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        new_fitted_params = {
            old_name: fitted_params[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_fitted_params.update(
            {
                param_name: param_data.fixed_to
                for param_name, param_data in self.fixed_params.items()
            }
        )

        new_fit_uncertainties = {
            old_name: fit_uncertainties[new_name]
            for new_name, old_name in self.param_mapping.items()
        }
        new_fit_uncertainties.update(
            {param_name: 0.0 for param_name, param_data in self.fixed_params.items()}
        )
        derived = self.model.calculate_derived_params(
            x=x,
            y=y,
            fitted_params=new_fitted_params,
            fit_uncertainties=new_fit_uncertainties,
        )
        mapping = self.derived_result_mapping
        for derived_dict in derived:
            for new_result_name, model_result_name in mapping.items():
                if new_result_name is None:
                    del derived_dict[model_result_name]
                    continue

                if model_result_name not in derived_dict:
                    raise ValueError(
                        f"Mapped derived result '{model_result_name}' not found."
                    )
                value = derived_dict.pop(model_result_name)
                if new_result_name in derived_dict:
                    raise ValueError(
                        f"Mapped derived result '{new_result_name}' duplicates "
                        "existing derived resut name."
                    )
                derived_dict[new_result_name] = value

        return derived



import logging
from typing import Callable, Dict, Optional, TYPE_CHECKING

import numpy as np
from scipy import stats
from statsmodels.stats import proportion


from .common import Model, TX, TY
from .MLE import MLEFitter
from .utils import Array

if TYPE_CHECKING:
    num_free_params = float

logger = logging.getLogger(__name__)


class BinomialFitter(MLEFitter):
    """Maximum-likelihood parameter estimator for Binomially-distributed data.

    The model is interpreted as giving the success probability for a Bernoulli
    trial under a given set of parameters: ``p = M(x; params)``.

    The y-axis data is interpreted as the success fraction, such that the total
    number of successes is equal to ``k = y * num_trails``.

    See :class:`~ionics_fits.common.Fitter` and :class:`~ionics_fits.MLE.MLEFitter` for
    further details.
    """

    TYPE: str = "Binomial"

    def __init__(
        self,
        x: TX,
        y: TY,
        num_trials: int,
        model: Model,
        step_size: float = 1e-4,
        minimizer_args: Optional[Dict] = None,
    ):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data
        :param y: y-axis data
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults. The
            model is (deep) copied and stored as an attribute.
        :param num_trials: number of Bernoulli trails for each sample
        :param step_size: see :class:`~ionics_fits.MLE.MLEFitter`.
        :param minimizer_args: optional dictionary of keyword arguments to be passed
            into ``scipy.optimize.minimize``.
        """
        self.num_trials = num_trials
        super().__init__(
            x=x, y=y, model=model, step_size=step_size, minimizer_args=minimizer_args
        )

    def log_likelihood(
        self,
        free_param_values: Array[("num_free_params",), np.float64],
        x: TX,
        y: TY,
        free_func: Callable[..., TY],
    ) -> float:
        p = free_func(x, *free_param_values.tolist())

        if np.any(p < 0) or np.any(p > 1):
            raise RuntimeError("Model values must lie between 0 and 1")

        n = self.num_trials
        k = np.rint(y * n, out=np.zeros_like(y, dtype=int), casting="unsafe")
        logP = stats.binom.logpmf(k=k, n=n, p=p)
        C = -np.sum(logP)

        return C

    def calc_sigma(self) -> TY:
        """Return an array of standard error values for each y-axis data point."""
        k = np.rint(
            self.y * self.num_trials,
            out=np.zeros_like(self.y, dtype=int),
            casting="unsafe",
        )

        lower, upper = proportion.proportion_confint(
            count=k,
            nobs=self.num_trials,
            alpha=1 - 0.6827,  # 1 sigma for Normal distributions
            method="beta",
        )

        sigma = 0.5 * (upper - lower)
        return sigma



from typing import Dict, List, Tuple
import numpy as np
from scipy.optimize import curve_fit

from . import heuristics
from .heuristics import get_spectrum
from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_power, scale_x, scale_y


class Gaussian(Model):
    """Gaussian model according to::

        y = a / (sigma * sqrt(2*pi)) * exp(-0.5*((x-x0)/(sigma))^2) + y0

    See :meth:`_func` for parameter details.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_x()),
        y0: ModelParameter(scale_y()),
        a: ModelParameter(scale_func=scale_power(x_power=1, y_power=1)),
        sigma: ModelParameter(lower_bound=0, scale_func=scale_x()),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param y0: y-axis offset
        :param a: y-axis scale factor. The Gaussian is normalized such that its integral
          is equal to ``a``.
        :param sigma: distribution half-width at ``1/e`` of maximum height is
          ``2*sigma`` (``sigma`` is the ``1/sqrt(e)`` radius).
        """
        y = (
            a / (sigma * np.sqrt(2 * np.pi)) * np.exp(-0.5 * ((x - x0) / sigma) ** 2)
            + y0
        )
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TX):
        x = np.squeeze(x)
        y = np.squeeze(y)

        basic_heuristics = self._estimate_parameters_basic(x=x, y=y)
        try:
            fft_heuristics = self._estimate_params_fft(x=x, y=y)
        except Exception:
            fft_heuristics = basic_heuristics
        try:
            peak_heuristics = self._estimate_params_peak(x=x, y=y)
        except Exception:
            peak_heuristics = basic_heuristics

        cost_fft = np.sum((y - self.func(x, fft_heuristics)) ** 2)
        cost_peak_heuristcs = np.sum((y - self.func(x, peak_heuristics)) ** 2)

        best = fft_heuristics if cost_fft < cost_peak_heuristcs else peak_heuristics

        for param_name, heuristic in best.items():
            self.parameters[param_name].heuristic = heuristic

    def _estimate_parameters_basic(self, x: TX, y: TY) -> Dict[str, float]:
        # fallback heuristics (e.g. don't fail even if we have a dataset that's too
        # small for the better heuristics)
        x0 = self.parameters["x0"].get_initial_value(np.mean(x))
        y0 = self.parameters["y0"].get_initial_value(0.5 * (y[0] + y[-1]))
        sigma = self.parameters["sigma"].get_initial_value(x.ptp() / 2)

        x0_ind = np.argmin(np.abs(x - x0))
        peak = y[x0_ind] - y0
        a = peak * sigma * np.sqrt(2 * np.pi)
        a = self.parameters["a"].get_initial_value(a)

        return {"a": a, "sigma": sigma, "y0": y0, "x0": x0}

    def _estimate_params_peak(self, x: TX, y: TY) -> Dict[str, float]:
        x0 = heuristics.get_sym_x(x=x, y=y)
        x0_ind = np.argmin(np.abs(x - x0))
        x0 = self.parameters["x0"].get_initial_value(x0)

        y0 = y[np.argmax(np.abs(x - x0))]
        y0 = self.parameters["y0"].get_initial_value(y0)

        peak = y[x0_ind] - y0
        inside = np.abs(y - y[x0_ind]) <= np.abs(peak * (1 - np.exp(-1)))
        full_width_1_e = x[inside].ptp()

        sigma = full_width_1_e / (2 * np.sqrt(2))
        sigma = self.parameters["sigma"].get_initial_value(sigma)

        a = peak * sigma * np.sqrt(2 * np.pi)
        a = self.parameters["a"].get_initial_value(a)

        return {"a": a, "sigma": sigma, "y0": y0, "x0": x0}

    def _estimate_params_fft(self, x: TX, y: TY) -> Dict[str, float]:
        # Gaussian Fourier Transform:
        #   F[A * exp(-(x/w)^2)](k) = A * sqrt(pi) * w * exp(-(pi*k*w)^2)
        #
        # This heuristic generally works extremely well when we have enough data but
        # struggles for smaller datasets

        fft_heuristics = {}
        y0 = self.parameters["y0"].get_initial_value(np.mean([y[0], y[-1]]))
        fft_heuristics["y0"] = y0

        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)

        def fun(omega, alpha, w):
            return alpha * np.exp(-((0.5 * omega * w) ** 2))

        alpha0 = abs_spectrum[0]
        w0 = 2 / omega[abs_spectrum < alpha0 * np.exp(-1)][0]
        (alpha, w), _ = curve_fit(
            f=fun, xdata=omega, ydata=abs_spectrum, p0=[alpha0, w0]
        )

        peak_idx = np.argmax(np.abs(y - y0))
        y_peak = y[peak_idx]
        sgn = 1 if y_peak > y0 else -1

        fft_heuristics["sigma"] = w / np.sqrt(2)
        fft_heuristics["a"] = sgn * alpha

        x0 = heuristics.find_x_offset_sym_peak_fft(
            model=self,
            x=x,
            y=y,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=4 / w,
            test_pts=x[peak_idx],
            defaults=fft_heuristics,
        )

        fft_heuristics["x0"] = x0

        return fft_heuristics

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Derived parameters:

        * ``FWHMH``: full width at half-maximum height
        * ``peak``: peak height above ``y0``
        * ``w0``: full width at ``1/e`` max height. For Gaussian beams this is the beam
          waist

        """

        sigma = fitted_params["sigma"]
        a = fitted_params["a"]

        derived_params = {}
        derived_params["FWHMH"] = 2.35482 * sigma
        derived_params["peak"] = a / (sigma * np.sqrt(2 * np.pi))
        derived_params["w0"] = 4 * sigma

        derived_uncertainties = {}
        derived_uncertainties["FWHMH"] = 2.35482 * fit_uncertainties["sigma"]
        derived_uncertainties["peak"] = (
            fit_uncertainties["a"] / (sigma * np.sqrt(2 * np.pi))
        ) ** 2 + (
            fit_uncertainties["sigma"] * a / (sigma**2 * np.sqrt(2 * np.pi))
        ) ** 2
        derived_uncertainties["w0"] = 4 * fit_uncertainties["sigma"]

        return derived_params, derived_uncertainties



import copy
import logging
import numpy as np
import pprint
from scipy import optimize
from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING

from .common import Fitter, Model, ModelParameter, TX, TY
from .utils import Array

if TYPE_CHECKING:
    num_free_params = float


logger = logging.getLogger(__name__)


class MLEFitter(Fitter):
    """Base class for maximum Likelihood Parameter Estimation fitters.

    Implementations should override the :meth:`log_likelihood` and
    :meth:`~ionics_fits.common.Fitter.calc_sigma` methods.

    See :class:`~ionics_fits.common.Fitter` for further details.
    """

    TYPE: str = "MLE"

    def __init__(
        self,
        x: TX,
        y: TY,
        model: Model,
        step_size: float = 1e-4,
        minimizer_args: Optional[Dict] = None,
    ):
        """Fits a model to a dataset and stores the results.

        :param x: x-axis data
        :param y: y-axis data
        :param model: the model function to fit to. The model's parameter dictionary is
            used to configure the fit (set parameter bounds etc). Modify this before
            fitting to change the fit behaviour from the model class' defaults. The
            model is (deep) copied and stored as an attribute.
        :param step_size: step size used when calculating the log likelihood's Hessian
            as part of finding the fitted parameter standard errors. Where finite
            parameter bounds are provided, they are used to scale the step size
            appropriately for each parameter.
        :param minimizer_args: optional dictionary of keyword arguments to be passed
            into ``scipy.optimize.minimize``.
        """
        self.minimizer_args = {"options": {"maxls": 100}}
        if minimizer_args is not None:
            self.minimizer_args.update(minimizer_args)

        self.step_size = step_size

        if np.any(y < 0) or np.any(y > 1):
            raise RuntimeError("y values must lie between 0 and 1")

        # Since we interpret the y-axis as a probability distribution, it should not be
        # rescaled
        def can_rescale() -> Tuple[List[bool], List[bool]]:
            rescale_xs, rescale_ys = self._can_rescale()
            return rescale_xs, [False] * len(rescale_ys)

        model = copy.deepcopy(model)
        self._can_rescale = model.can_rescale
        model.can_rescale = can_rescale

        super().__init__(x=x, y=y, model=model)

    def log_likelihood(
        self,
        free_param_values: Array[("num_free_params",), np.float64],
        x: TX,
        y: TY,
        free_func: Callable[..., TY],
    ) -> float:
        """Returns the negative log-likelihood of a given dataset

        :param free_param_values: array of floated parameter values
        :param x: x-axis data
        :param y: y-axis data
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters
        """
        raise NotImplementedError

    def _fit(
        self,
        x: TX,
        y: TY,
        parameters: Dict[str, ModelParameter],
        free_func: Callable[..., TY],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """Performs maximum likelihood parameter estimation and calculates standard
        errors in each parameter.

        :param x: rescaled x-axis data, must be a 1D array
        :param y: rescaled y-axis data
        :param parameters: dictionary of rescaled model parameters
        :param free_func: convenience wrapper for the model function, taking only values
            for the fit's free parameters

        :returns: tuple of dictionaries mapping model parameter names to their fitted
            values and uncertainties.
        """
        free_parameters = [
            param_name
            for param_name, param_data in parameters.items()
            if param_data.fixed_to is None
        ]
        num_free_params = len(free_parameters)

        p0 = [parameters[param].get_initial_value() for param in free_parameters]
        lower = [parameters[param].lower_bound for param in free_parameters]
        upper = [parameters[param].upper_bound for param in free_parameters]
        p0_dict = {
            param: parameters[param].get_initial_value() for param in free_parameters
        }

        logger.debug(
            f"Starting {self.TYPE} fitting with initial parameters: "
            f"{pprint.pformat(p0_dict)}"
        )

        # maxls setting helps with ABNORMAL_TERMINATION_IN_LNSRCH
        res = optimize.minimize(
            fun=self.log_likelihood,
            args=(x, y, free_func),
            x0=p0,
            bounds=zip(lower, upper),
            **self.minimizer_args,
        )

        if not res.success:
            raise RuntimeError(f"{self.TYPE} fit failed: {res.message}")

        # Compute parameter covariance matrix
        #
        # The covariance matrix is given by the inverse of the Hessian at the optimum.
        #
        # While scipy.minimize provides an approximate value for the inverse
        # Hessian, it's not accurate enough to be used for error estimation.
        # we perform our own calculation based on finite differences using the
        # specified step size, rescaled by the parameter bounds where possible.
        lower = [bound if bound is not None else -np.inf for bound in lower]
        upper = [bound if bound is not None else +np.inf for bound in upper]

        def diff(param_idx, fun):
            if np.isfinite(lower[param_idx]) and np.isfinite(upper[param_idx]):
                param_range = upper[param_idx] - lower[param_idx]
                step_size = self.step_size * param_range
            else:
                step_size = self.step_size

            def _fun(free_param_values):
                param_value = free_param_values[param_idx]
                param_upper = np.clip(
                    param_value + 0.5 * step_size,
                    a_min=lower[param_idx],
                    a_max=upper[param_idx],
                )
                param_lower = np.clip(
                    param_value - 0.5 * step_size,
                    a_min=lower[param_idx],
                    a_max=upper[param_idx],
                )
                delta = param_upper - param_lower

                param_upper_values = np.copy(free_param_values)
                param_lower_values = np.copy(free_param_values)
                param_upper_values[param_idx] = param_upper
                param_lower_values[param_idx] = param_lower

                f_upper = fun(param_upper_values)
                f_lower = fun(param_lower_values)

                return (f_upper - f_lower) / delta

            return _fun

        def log_likelihood(free_param_values):
            return self.log_likelihood(
                free_param_values=free_param_values, x=x, y=y, free_func=free_func
            )

        hessian = np.zeros((num_free_params, num_free_params))

        for i_idx in range(num_free_params):
            for j_idx in range(num_free_params):
                first_diff = diff(i_idx, log_likelihood)
                second_diff = diff(j_idx, first_diff)
                hessian[i_idx, j_idx] = second_diff(res.x)

        p_cov = np.linalg.inv(hessian)
        p_err = np.sqrt(np.diag(p_cov))

        p = {param: value for param, value in zip(free_parameters, res.x)}
        p_err = {param: value for param, value in zip(free_parameters, p_err)}

        return p, p_err



from typing import Dict, List, Tuple

import numpy as np

from . import heuristics
from .utils import PeriodicModelParameter
from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_invariant, scale_x, scale_x_inv, scale_y


class Ramsey(Model):
    r"""Fit model for detuning scans of Ramsey experiments (for time scans, use the
    Sinusoid model).

    This model calculates the measurement outcomes for Ramsey experiments, defined by::

        P = P_readout_g + (P_readout_e - P_readout_g) * P_e

    where ``P_e`` is the (time-dependent) population in the excited state and
    ``P_readout_g`` and ``P_readout_e`` are the readout levels (measurement outcomes
    when the qubit is in one state).

    The model requires that the system starts out entirely in one of the ground or
    excited states, specified using :meth:``__init__``\'s ``start_excited`` parameter.

    All frequencies are in angular units.
    """

    def __init__(self, start_excited: bool):
        super().__init__()
        self.start_excited = start_excited

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [False]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        P_readout_e: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        P_readout_g: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        t: ModelParameter(
            lower_bound=0.0,
            scale_func=scale_x_inv(),
        ),
        t_pi_2: ModelParameter(
            lower_bound=0.0,
            scale_func=scale_x_inv(),
        ),
        w_0: ModelParameter(scale_func=scale_x()),
        phi: PeriodicModelParameter(
            period=2 * np.pi,
            offset=-np.pi,
            scale_func=scale_invariant,
        ),
        tau: ModelParameter(
            lower_bound=0.0,
            fixed_to=np.inf,
            scale_func=scale_x_inv(),
        ),
    ):
        """
        :param P_readout_e: excited state readout level
        :param P_readout_g: ground state readout level
        :param t: Ramsey delay
        :param t_pi_2: duration of the pi/2 pulses. The pi/2 pulses are assumed to be
            ideal pi/2 pulses with a corresponding Rabi frequency of
            ``Omega = np.pi / (2 * t_pi_2)``
        :param w_0: resonance frequency offset, defined such that the Ramsey detuning is
            given by ``delta = x - w_0``
        :param phi: phase of the second pi/2 pulse relative to the first pi/2 pulse
        :param tau: decay time constant (fixed to infinity by default)
        """
        delta = x - w_0
        Omega = np.pi / (2 * t_pi_2)

        alpha = (delta / Omega) ** 2
        theta = np.pi / 4 * np.sqrt(1 + alpha)
        gamma = np.arctan(np.sqrt(alpha / (1 + alpha)) * np.tan(theta))

        P_trans = (
            2
            * np.exp(-t / tau)
            / (1 + alpha)
            * np.sin(theta) ** 2
            * (np.cos(theta) ** 2 + alpha / (1 + alpha) * np.sin(theta) ** 2)
            * (1 + np.cos(delta * t + phi + 2 * gamma))
        )

        P_e = 1 - P_trans if self.start_excited else P_trans
        return P_readout_g + (P_readout_e - P_readout_g) * P_e

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        if self.start_excited:
            self.parameters["P_readout_e"].heuristic = y[0]
            self.parameters["P_readout_g"].heuristic = abs(1 - y[0])
        else:
            self.parameters["P_readout_g"].heuristic = y[0]
            self.parameters["P_readout_e"].heuristic = abs(1 - y[0])

        unknowns = {
            param_name
            for param_name, param_data in self.parameters.items()
            if not param_data.has_user_initial_value() and param_name != "tau"
        }

        # This is the most common use-case for this model. TODO: cover other use-cases
        # as they arise...
        if unknowns.issubset({"w_0", "phi"}):
            self.parameters["phi"].heuristic = 0.0
            self.parameters["w_0"].heuristic = heuristics.find_x_offset_sampling(
                model=self,
                x=x,
                y=y,
                width=2 * np.pi / self.parameters["t"].get_initial_value(),
                x_offset_param_name="w_0",
            )
        else:
            raise ValueError(
                f"No Ramsey heuristic currently available for the unknowns: {unknowns}"
            )

        self.parameters["tau"].heuristic = 10 * self.parameters["t"].get_initial_value()

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Derived parameters:

        * ``f_0``: resonance frequency offset in linear units, given by
          ``w_0 / (2 * np.pi)``
        """
        derived_params = {}
        derived_uncertainties = {}

        derived_params["f_0"] = fitted_params["w_0"] / (2 * np.pi)
        derived_uncertainties["f_0"] = fit_uncertainties["w_0"] / (2 * np.pi)

        return derived_params, derived_uncertainties



from typing import Dict, List, Tuple

import numpy as np

from . import heuristics
from .sinc import Sinc2
from .sinusoid import Sinusoid
from .heuristics import get_spectrum
from ..common import Model, ModelParameter, TX, TY
from ..normal import NormalFitter
from ..utils import scale_undefined, scale_x, scale_x_inv, scale_y


class RabiFlop(Model):
    r"""Base class for damped Rabi flopping.

    This model calculates measurement outcomes for two-state systems undergoing damped
    Rabi oscillations, defined by::

        P = P_readout_g + (P_readout_e - P_readout_g) * P_e

    where ``P_e`` is the (time-dependent) population in the excited state and
    ``P_readout_g`` and ``P_readout_e`` are the readout levels (measurement outcomes
    when the qubit is in one state).

    This class does not support fitting directly; use one of the subclasses instead.

    The model requires that the system starts out entirely in one of the ground or
    excited states, specified using :meth:`__init__`\'s
    ``start_excited`` parameter.

    The probability of transition from one state to the other is calculated as::

        P_trans = 1 / 2 * omega^2 / W^2 * [1 - exp(-t / tau) * cos(W * t)]

    where:
        * ``t`` is the duration of interaction between qubit and driving field
        * ``W = sqrt(omega^2 + delta^2)``
        * ``delta`` is the detuning of the driving field from resonance
        * ``omega`` is the Rabi frequency
        * ``tau`` is the decay time constant.

    Independent variables:
        * ``t_pulse``: duration of driving pulse including dead time. The duration of
          the interaction is given by ``t = max(0, t_pulse - t_dead)``.
        * ``w``: frequency of driving pulse relative to the reference frequency ``w_0``,
          given by ``delta = w - w_0``

    All frequencies are in angular units.
    """

    def __init__(self, start_excited: bool):
        """
        :param start_excited: if ``True`` the system is assumed to start in the excited
            state, other wise it is assumed to start in the ground state.
        """
        super().__init__()
        self.start_excited = start_excited

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [False]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        P_readout_e: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        P_readout_g: ModelParameter(
            lower_bound=0.0,
            upper_bound=1.0,
            scale_func=scale_y(),
        ),
        omega: ModelParameter(lower_bound=0.0, scale_func=scale_undefined),
        tau: ModelParameter(
            lower_bound=0.0, fixed_to=np.inf, scale_func=scale_undefined
        ),
        t_dead: ModelParameter(
            lower_bound=0.0, fixed_to=0.0, scale_func=scale_undefined
        ),
        w_0: ModelParameter(scale_func=scale_undefined),
    ) -> TY:
        """Return measurement probability.

        :param x: tuple of ``(t_pulse, w)``. Subclasses should override func to
            map this onto the appropriate input data.
        :param P_readout_e: excited state readout level
        :param P_readout_g: ground state readout level
        :param omega: Rabi frequency
        :param tau: decay time constant (fixed to infinity by default)
        :param t_dead: dead time (fixed to 0 by default)
        :param w_0: resonance frequency offset
        """
        t = np.clip(x[0] - t_dead, a_min=0.0, a_max=None)
        delta = x[1] - w_0
        W = np.sqrt(omega**2 + delta**2)

        P_trans = (
            0.5
            * np.divide(omega**2, W**2, out=np.zeros_like(W), where=(W != 0.0))
            * (1 - np.exp(-t / tau) * np.cos(W * t))
        )
        P_e = 1 - P_trans if self.start_excited else P_trans

        return P_readout_g + (P_readout_e - P_readout_g) * P_e

    # pytype: enable=invalid-annotation

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        Derived parameters:
            - ``t_pi``: Pi-time, calculated as ``t_pi = pi / omega``
            - ``t_pi_2``: Pi/2-time, calculated as ``t_pi_2 = t_pi / 2``
            - ``f_0``: Offset of resonance from zero of frequency variable in linear
              units
        """
        omega = fitted_params["omega"]
        t_pi = np.pi / omega
        t_pi_2 = t_pi / 2

        omega_err = fit_uncertainties["omega"]

        derived_params = {}
        derived_params["t_pi"] = t_pi
        derived_params["t_pi_2"] = t_pi_2

        derived_uncertainties = {}
        derived_uncertainties["t_pi"] = omega_err * np.pi / (omega**2)
        derived_uncertainties["t_pi_2"] = derived_uncertainties["t_pi"] / 2

        if "w_0" in fitted_params:
            derived_params["f_0"] = fitted_params["w_0"] / (2 * np.pi)
            derived_uncertainties["f_0"] = fit_uncertainties["w_0"] / (2 * np.pi)

        return derived_params, derived_uncertainties


class RabiFlopFreq(RabiFlop):
    """Fit model for Rabi flopping frequency scans.

    This model calculates the measurement outcomes for damped Rabi flops when the
    pulse duration is kept fixed and only its frequency is varied. The pulse duration is
    specified using a new ``t_pulse`` model parameter.

    See also :class:`~ionics_fits.models.rabi.RabiFlop`.
    """

    def __init__(self, start_excited: bool):
        super().__init__(start_excited)

        self.parameters["t_pulse"] = ModelParameter(
            lower_bound=0.0, scale_func=scale_x_inv()
        )

        self.parameters["omega"].scale_func = scale_x()
        self.parameters["w_0"].scale_func = scale_x()

        self.parameters["tau"].scale_func = scale_x_inv()
        self.parameters["t_dead"].scale_func = scale_x_inv()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        param_values = param_values.copy()
        t_pulse = param_values.pop("t_pulse")
        return self._func(
            (t_pulse, x), **param_values
        )  # pytype: disable=wrong-arg-types

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        self.parameters["t_dead"].heuristic = 0.0

        if self.start_excited:
            self.parameters["P_readout_e"].heuristic = y[0]
            self.parameters["P_readout_g"].heuristic = abs(1 - y[0])
        else:
            self.parameters["P_readout_g"].heuristic = y[0]
            self.parameters["P_readout_e"].heuristic = abs(1 - y[0])

        y0_param = "P_readout_e" if self.start_excited else "P_readout_g"

        # A common use of this model is finding `w_0` when the Rabi frequency and pulse
        # duration are known. In this case we don't need to rely on the Sinc2
        # approximation
        unknowns = set()
        for param, param_data in self.parameters.items():
            try:
                param_data.get_initial_value()
            except ValueError:
                unknowns.add(param)

        if unknowns == {"w_0"}:
            omega, spectrum = get_spectrum(x, y, trim_dc=True)
            w_0 = heuristics.find_x_offset_sym_peak_fft(
                model=self,
                x=x,
                y=y,
                omega=omega,
                spectrum=spectrum,
                omega_cut_off=self.parameters["t_pulse"].get_initial_value(),
                x_offset_param_name="w_0",
                y_offset_param_name=y0_param,
            )
            self.parameters["w_0"].heuristic = w_0
            return

        # There isn't a simple analytic form for the Fourier transform of a Rabi
        # flop in the general case. However in the low pulse area limit (and
        # ignoring decay etc) the Rabi flop function tends to the sinc^2 function:
        #   (omega * t / 2)^2 * sinc^2(delta * t / 2)
        # NB np.sinc(x) = np.sin(pi * x) / (pi * x)
        # This heuristic breaks down when: omega * t_pulse ~ pi
        model = Sinc2()
        y0 = self.parameters[y0_param].get_initial_value()
        model.parameters["y0"].fixed_to = y0
        fit = NormalFitter(x, y, model)

        self.parameters["t_pulse"].heuristic = 2 * fit.values["w"]
        t_pulse = self.parameters["t_pulse"].get_initial_value()
        self.parameters["omega"].heuristic = (
            2 * np.sqrt(np.abs(fit.values["a"])) / t_pulse
        )

        if self.parameters["w_0"].has_user_initial_value():
            return

        # The user hasn't told us what w_0 is so we need to find a heuristic value
        # In addition to going off the Sinc^2, we use a simple sampling-based heuristic
        x_sinc = fit.values["x0"]

        # Test out all points with a contrast of 30% of more. NB the fitter
        # automatically rescales our y-data so this assumes we have one point at
        # sufficiently high contrast for the y-axis rescaling to not do much!
        x_sample = x[np.argwhere(np.abs(y - y0) > 0.3)]
        x_trial = np.append(x_sample, [x_sinc])
        w_0, _ = heuristics.param_min_sqrs(
            model=self, x=x, y=y, scanned_param="w_0", scanned_param_values=x_trial
        )
        self.parameters["w_0"].heuristic = w_0

        self.parameters["tau"].heuristic = 10 * t_pulse


class RabiFlopTime(RabiFlop):
    """Fit model for Rabi flopping pulse duration scans.

    This model calculates the measurement outcome for damped Rabi flops when the
    frequency of the pulse is kept fixed and only its duration is varied.

    Since the detuning is not scanned as an independent variable, we replace ``w_0``
    with a new model parameter ``delta``, defined by: ``delta = |w - w_0|``.

    See also :class:`~ionics_fits.models.rabi.RabiFlop`.
    """

    def __init__(self, start_excited: bool):
        super().__init__(start_excited)

        self.parameters["delta"] = ModelParameter(scale_func=scale_x_inv())
        del self.parameters["w_0"]

        self.parameters["omega"].scale_func = scale_x_inv()

        self.parameters["tau"].scale_func = scale_x()
        self.parameters["t_dead"].scale_func = scale_x()

    def func(self, x: TX, param_values: Dict[str, float]) -> TY:
        param_values = param_values.copy()
        delta = param_values.pop("delta")
        param_values["w_0"] = 0.0
        return self._func((x, delta), **param_values)  # pytype: disable=wrong-arg-types

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        self.parameters["t_dead"].heuristic = 0.0
        self.parameters["tau"].heuristic = 10 * x.ptp()

        if self.start_excited:
            self.parameters["P_readout_e"].heuristic = y[0]
            self.parameters["P_readout_g"].heuristic = abs(1 - y[0])
        else:
            self.parameters["P_readout_g"].heuristic = y[0]
            self.parameters["P_readout_e"].heuristic = abs(1 - y[0])

        P_readout_e = self.parameters["P_readout_e"].get_initial_value()
        P_readout_g = self.parameters["P_readout_g"].get_initial_value()

        model = Sinusoid()
        if P_readout_e >= P_readout_g:
            model.parameters["phi"].fixed_to = (
                np.pi / 2 if self.start_excited else 3 * np.pi / 2
            )
        else:
            model.parameters["phi"].fixed_to = (
                3 * np.pi / 2 if self.start_excited else np.pi / 2
            )

        fit = NormalFitter(x, y, model)

        W = fit.values["omega"]
        self.parameters["omega"].heuristic = np.sqrt(2 * fit.values["a"]) * W
        omega = self.parameters["omega"].get_initial_value()

        if W >= omega:
            self.parameters["delta"].heuristic = np.sqrt(W**2 - omega**2)
        else:
            # can't use param_min_sqrs because omega and delta are coupled
            deltas = np.linspace(0, W / 2, 10)
            omegas = np.sqrt(W**2 - deltas**2)
            costs = np.zeros_like(deltas)

            initial_values = {
                param: param_data.get_initial_value()
                for param, param_data in self.parameters.items()
                if param != "delta"
            }

            for idx in range(len(deltas)):
                initial_values["delta"] = deltas[idx]
                initial_values["omega"] = omegas[idx]
                y_idx = self.func(x, initial_values)
                costs[idx] = np.sqrt(np.sum((y - y_idx) ** 2))
            opt_idx = np.argmin(costs)

            self.parameters["delta"].heuristic = deltas[opt_idx]
            self.parameters["omega"].heuristic = omegas[opt_idx]

        # Corner-case: if the time axis starts from t_0 >> t_pi the above heuristic
        # can fail. This is because the accuracy of the sinusoid fit is limited by the
        # step size in frequency space. Once the uncertainty in the Rabi frequency
        # estimate becomes such that we don't know if we've done n or (n + 1) flops
        # before t_0 the fits will start failing.
        d_omega = 2 * np.pi / x.ptp()  # approx uncertainty in Rabi freq from FFT
        t_pi = np.pi / self.parameters["omega"].get_initial_value()
        d_t_pi = np.pi / d_omega
        n_pi = min(x) / t_pi
        err = d_t_pi * n_pi  # number of t_pi worth of uncertainty in the fit

        if err > 0.1:
            n_pi_min = max(n_pi - 4 * err, 0.0)
            n_pi_max = n_pi + 4 * err
            num_pts = int(np.rint(20 * (n_pi_max - n_pi_min)))

            omega_min = np.pi * n_pi_min / min(x)
            omega_max = np.pi * n_pi_max / min(x)
            omegas = np.linspace(omega_min, omega_max, num_pts)

            self.parameters["omega"].heuristic, _ = heuristics.param_min_sqrs(
                model=self,
                x=x,
                y=y,
                scanned_param="omega",
                scanned_param_values=omegas,
            )



import numpy as np
from typing import Dict, List, Tuple

from . import heuristics
from ..common import Model, ModelParameter, TX, TY
from .transformations.reparametrized_model import ReparametrizedModel
from ..utils import scale_x, scale_x_inv, scale_invariant, scale_y
from . import utils


class Sinusoid(Model):
    """Generalised sinusoid fit according to::

        y = Gamma * a * sin[omega * (x - x0) + phi] + y0
        Gamma = exp(-x / tau).

    All phases are in radians, frequencies are in angular units.

    ``x0`` and ``phi0`` are equivalent parametrisations for the phase offset, but in
    some cases it works out convenient to have access to both (e.g. one as a fixed
    offset, the other floated). At most one of them should be floated at once. By
    default, ``x0`` is fixed at 0 and ``phi0`` is floated.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        a: ModelParameter(lower_bound=0, scale_func=scale_y()),
        omega: ModelParameter(lower_bound=0, scale_func=scale_x_inv()),
        phi: utils.PeriodicModelParameter(
            period=2 * np.pi,
            offset=-np.pi,
            scale_func=scale_invariant,
        ),
        y0: ModelParameter(scale_func=scale_y()),
        x0: ModelParameter(fixed_to=0, scale_func=scale_x()),
        tau: ModelParameter(
            lower_bound=0,
            fixed_to=np.inf,
            scale_func=scale_x(),
        ),
    ) -> TY:
        """
        :param a: initial (``x = 0``) amplitude of the sinusoid
        :param omega: angular frequency
        :param phi: phase offset
        :param y0: y-axis offset
        :param x0: x-axis offset
        :param tau: decay/growth constant
        """
        Gamma = np.exp(-x / tau)
        return Gamma * a * np.sin(omega * (x - x0) + phi) + y0

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        # We don't have good heuristics for these parameters
        self.parameters["y0"].heuristic = np.mean(y)
        self.parameters["tau"].heuristic = np.max(x)

        omega, spectrum = heuristics.get_pgram(x, y)
        peak = np.argmax(spectrum)

        self.parameters["a"].heuristic = spectrum[peak]
        self.parameters["omega"].heuristic = omega[peak]

        phi, _ = heuristics.param_min_sqrs(
            model=self,
            x=x,
            y=y,
            scanned_param="phi",
            scanned_param_values=np.linspace(-np.pi, np.pi, num=20),
            defaults={"x0": 0},
        )

        phi = self.parameters["phi"].clip(phi)

        if self.parameters["x0"].fixed_to is None:
            if self.parameters["phi"].fixed_to is None:
                raise ValueError("Only one of 'x0' and 'phi' may be floated at once")

            self.parameters["phi"].heuristic = 0
            self.parameters["x0"].heuristic = (
                -phi / self.parameters["omega"].get_initial_value()
            )
        else:
            self.parameters["phi"].heuristic = phi
            self.parameters["x0"].heuristic = 0.0

    def calculate_derived_params(
        self,
        x: TX,
        y: TY,
        fitted_params: Dict[str, float],
        fit_uncertainties: Dict[str, float],
    ) -> Tuple[Dict[str, float], Dict[str, float]]:
        """
        * ``f``: frequency
        * ``phi_cosine``: cosine phase (``phi + pi/2``)
        * ``contrast``: peak-to-peak amplitude of the pure sinusoid
        * ``min``/``max``: min / max values of the pure sinusoid
        * ``period``: period of oscillation

        TODO: peak values of the damped sinusoid as well as ``x`` value that the peak
        occurs at.
        """
        derived_params = {}
        derived_params["f"] = fitted_params["omega"] / (2 * np.pi)
        derived_params["phi_cosine"] = fitted_params["phi"] + np.pi / 2
        derived_params["contrast"] = 2 * np.abs(fitted_params["a"])
        derived_params["min"] = fitted_params["y0"] - np.abs(fitted_params["a"])
        derived_params["max"] = fitted_params["y0"] + np.abs(fitted_params["a"])
        derived_params["period"] = 2 * np.pi / fitted_params["omega"]

        derived_uncertainties = {}
        derived_uncertainties["f"] = fit_uncertainties["omega"] / (2 * np.pi)
        derived_uncertainties["phi_cosine"] = fit_uncertainties["phi"]
        derived_uncertainties["contrast"] = 2 * fit_uncertainties["a"]
        derived_uncertainties["min"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["max"] = np.sqrt(
            fit_uncertainties["y0"] ** 2 + fit_uncertainties["a"] ** 2
        )
        derived_uncertainties["period"] = (
            2 * np.pi * fit_uncertainties["omega"] / (fitted_params["omega"] ** 2)
        )

        return derived_params, derived_uncertainties


class SineMinMax(ReparametrizedModel):
    """Sinusoid parametrised by minimum / maximum values instead of offset / amplitude::

            y = Gamma * a * sin[omega * (x - x0) + phi] + y0

    This class is equivalent to :class:`Sinusoid` except that the ``a`` and ``y0``
    parameters are replaced with new ``min`` and ``max`` parameters defined by::

      min = y0 - a
      max = y0 + a

    See :class:`Sinusoid` for further details.
    """

    def __init__(self):
        super().__init__(
            model=Sinusoid(),
            new_params={
                "min": ModelParameter(scale_func=scale_y()),
                "max": ModelParameter(scale_func=scale_y()),
            },
            bound_params=["a", "y0"],
        )

    @staticmethod
    def bound_param_values(param_values: Dict[str, float]) -> Dict[str, float]:
        return {
            "a": 0.5 * (param_values["max"] - param_values["min"]),
            "y0": 0.5 * (param_values["max"] + param_values["min"]),
        }

    @staticmethod
    def bound_param_uncertainties(
        param_values: Dict[str, float], param_uncertainties: Dict[str, float]
    ) -> Dict[str, float]:
        err = 0.5 * np.sqrt(
            param_uncertainties["max"] ** 2 + param_uncertainties["min"] ** 2
        )
        return {"a": err, "y0": err}

    @staticmethod
    def new_param_values(model_param_values: Dict[str, float]) -> Dict[str, float]:
        return {
            "max": (model_param_values["y0"] + model_param_values["a"]),
            "min": (model_param_values["y0"] - model_param_values["a"]),
        }



from typing import Tuple, List
import numpy as np

from . import heuristics
from .heuristics import get_spectrum
from ..common import Model, ModelParameter, TX, TY
from ..utils import scale_x, scale_y


class Lorentzian(Model):
    """Lorentzian model according to::

        y = a * fwhmh^2 / ((x - x0)^2 + fwhmh^2) + y0

    See :meth:`_func` for parameter details.
    """

    def get_num_x_axes(self) -> int:
        return 1

    def get_num_y_axes(self) -> int:
        return 1

    def can_rescale(self) -> Tuple[List[bool], List[bool]]:
        return [True], [True]

    # pytype: disable=invalid-annotation
    def _func(
        self,
        x: TX,
        x0: ModelParameter(scale_func=scale_x()),
        y0: ModelParameter(scale_func=scale_y()),
        a: ModelParameter(scale_func=scale_y()),
        fwhmh: ModelParameter(lower_bound=0, scale_func=scale_x()),
    ) -> TY:
        """
        :param x0: x-axis offset
        :param y0: y-axis offset
        :param a: peak value of the function above ``y0``
        :param fwhmh: full width at half maximum height of the function
        """
        y = a * fwhmh**2 / ((x - x0) ** 2 + fwhmh**2) + y0
        return y

    # pytype: enable=invalid-annotation

    def estimate_parameters(self, x: TX, y: TY):
        x = np.squeeze(x)
        y = np.squeeze(y)

        # Fourier transform:
        #  f(y) = a * fwhmh^2 / ((x - x0)^2 + fwhmh^2) + y0
        #  f(k) = a * fwhmh * pi * exp(-2*pi*i*k*x0) * exp(-2*pi*fwhmh*|k|)
        omega, spectrum = get_spectrum(x, y, trim_dc=True)
        abs_spectrum = np.abs(spectrum)
        k = omega / (2 * np.pi)

        peak = abs_spectrum[0]
        W = peak * np.exp(-1)

        idx_1_e = np.argmin(np.abs(abs_spectrum - W))

        # We usually don't have great spectral resolution around the peak so interpolate
        if k[idx_1_e] > W:
            upper_idx = idx_1_e
            lower_idx = idx_1_e + 1
        else:
            upper_idx = idx_1_e - 1
            lower_idx = idx_1_e

        # if we don't have enough data to figure this out, set the half-width to one
        # sample wide
        df_dk = (abs_spectrum[lower_idx] - abs_spectrum[upper_idx]) / (k[1] - k[0])
        if df_dk == 0:
            tau = k[1] - k[0]
        else:
            df = abs_spectrum[idx_1_e] - W
            tau = k[idx_1_e] - df / df_dk

        fwhmh = 1 / (2 * np.pi * tau)
        a = peak / (np.pi * fwhmh)

        self.parameters["y0"].heuristic = np.mean([y[0], y[-1]])
        y0 = self.parameters["y0"].get_initial_value()
        peak_idx = np.argmax(np.abs(y - y0))
        y_peak = y[peak_idx]
        sgn = 1 if y_peak > y0 else -1

        self.parameters["a"].heuristic = a * sgn
        self.parameters["fwhmh"].heuristic = fwhmh

        cut_off = 2 * tau

        x0 = heuristics.find_x_offset_sym_peak_fft(
            model=self,
            x=x,
            y=y,
            omega=omega,
            spectrum=spectrum,
            omega_cut_off=cut_off,
            test_pts=x[peak_idx],
        )

        self.parameters["x0"].heuristic = x0
